from marshmallow import fields, RAISE

from argrelay.enum_desc.DistinctValuesQuery import DistinctValuesQuery
from argrelay.misc_helper_common.ObjectSchema import ObjectSchema
from argrelay.misc_helper_common.TypeDesc import TypeDesc
from argrelay.mongo_data.MongoConfig import MongoConfig
from argrelay.schema_config_core_server.MongoClientConfigSchema import mongo_client_config_desc
from argrelay.schema_config_core_server.MongoServerConfigSchema import mongo_server_config_desc

use_mongomock_ = "use_mongomock"
distinct_values_query_ = "distinct_values_query"
mongo_client_ = "mongo_client"
mongo_server_ = "mongo_server"


class MongoConfigSchema(ObjectSchema):
    class Meta:
        unknown = RAISE
        strict = True

    model_class = MongoConfig

    use_mongomock = fields.Boolean()
    """
    It might be the case that (test) `mongomock` lib actually provides necessary functionality and meet
    non-function requirements for many (prod) workloads without a need to deploy and administer MongoDB.
    https://github.com/mongomock/mongomock
    """

    distinct_values_query = fields.Enum(
        DistinctValuesQuery,
        required = True,
    )
    """
    See `DistinctValuesQuery`.
    """

    mongo_client = fields.Nested(mongo_client_config_desc.dict_schema)

    mongo_server = fields.Nested(mongo_server_config_desc.dict_schema)


mongo_config_desc = TypeDesc(
    dict_schema = MongoConfigSchema(),
    ref_name = MongoConfigSchema.__name__,
    dict_example = {
        use_mongomock_: True,
        distinct_values_query_: DistinctValuesQuery.original_find_and_loop.name,
        mongo_client_: mongo_client_config_desc.dict_example,
        mongo_server_: mongo_server_config_desc.dict_example,
    },
    default_file_path = "",
)



import signal
import socket
import subprocess
import time
from contextlib import closing
from datetime import datetime, timedelta

from argrelay.misc_helper_common import get_argrelay_dir
from argrelay.runtime_data.ClientConfig import ClientConfig
from argrelay.schema_config_core_client.ClientConfigSchema import client_config_desc
from argrelay.test_infra import change_to_known_repo_path
from argrelay.test_infra.InOutTestClass import InOutTestClass


class ClientServerTestClass(InOutTestClass):
    """
    Unlike tests derived from `ManualServerTestClass` (which rely on the server to be started manually),
    the test controls startup of the server.

    This test starts server (only) via their generated files in `@/bin/` dir.

    Avoiding generated file in `@/bin/` dir for client
    allows executing client code within the OS process responsible for running this test.
    In turn, this allows using `EnvMockBuilder` for client-side mocking to
    intercept server responses (see `RemoteTestClass`).
    """

    server_proc = None

    @classmethod
    def setUpClass(cls):
        InOutTestClass.setUpClass()
        cls.start_server()

    @classmethod
    def tearDownClass(cls):
        InOutTestClass.setUpClass()
        cls.stop_server()

    @classmethod
    def start_server(cls):
        with change_to_known_repo_path("."):
            cls.server_proc = subprocess.Popen([
                f"{get_argrelay_dir()}/exe/run_argrelay_server",
            ])
        cls.wait_for_connection_to_server()

    @classmethod
    def stop_server(cls):
        # shutdown gracefully:
        cls.server_proc.send_signal(signal.SIGINT)
        cls.server_proc.communicate()
        cls().assertEqual(0, cls.server_proc.returncode)

    @staticmethod
    def wait_for_connection_to_server():
        client_config: ClientConfig = client_config_desc.obj_from_default_file()

        delay_s = 1
        # 10 mins should be enough, right?
        timeout_ts = datetime.now() + timedelta(minutes = 10.0)
        with closing(socket.socket(socket.AF_INET, socket.SOCK_STREAM)) as sock:
            while True:
                result = sock.connect_ex((
                    client_config.connection_config.server_host_name,
                    client_config.connection_config.server_port_number,
                ))
                if result == 0:
                    print("connected to server")
                    return
                else:
                    now_ts = datetime.now()
                    if now_ts > timeout_ts:
                        raise TimeoutError
                    print(
                        f"now: {now_ts.time()} timeout: {timeout_ts.time()} - waiting {delay_s} sec for connection to server..."
                    )
                    time.sleep(delay_s)
                    delay_s += 1 if delay_s < 30 else 0
                    continue



import json

from argrelay.client_pipeline.BytesHandlerAbstract import BytesHandlerAbstract
from argrelay.handler_response.ClientResponseHandlerAbstract import ClientResponseHandlerAbstract
from argrelay.misc_helper_common.ElapsedTime import ElapsedTime


class BytesHandlerJson(BytesHandlerAbstract):

    def __init__(
        self,
        client_response_handler: ClientResponseHandlerAbstract,
    ):
        super().__init__(
        )
        self.client_response_handler = client_response_handler

    def handle_bytes(
        self,
        given_bytes: bytes,
    ):
        # Leave both object creation and validation via schemas to `client_response_handler`.
        # Just deserialize into dict here:
        response_dict = json.loads(given_bytes.decode())
        ElapsedTime.measure("after_deserialization")
        self.client_response_handler.handle_response(response_dict)



from __future__ import annotations

from copy import deepcopy

from argrelay.enum_desc.PluginType import PluginType
from argrelay.plugin_interp.AbstractInterp import AbstractInterp
from argrelay.runtime_context.AbstractPluginServer import AbstractPluginServer
from argrelay.runtime_context.InterpContext import InterpContext
from argrelay.runtime_data.ServerConfig import ServerConfig


class AbstractInterpFactory(AbstractPluginServer):

    def __init__(
        self,
        server_config: ServerConfig,
        plugin_instance_id: str,
        plugin_config_dict: dict,
    ):
        super().__init__(
            server_config,
            plugin_instance_id,
            plugin_config_dict,
        )

        self.interp_tree_abs_paths: list[tuple[str, ...]] = []
        """
        All tree paths this plugin is plugged into.

        It is populated via `load_interp_tree_abs_paths`.
        """

        # Takes part in implementation of FS_01_89_09_24 interp tree:
        self.interp_tree_abs_paths_to_node_configs: dict[tuple[str, ...], dict] = {}
        """
        Classes derived from `AbstractInterp` (which are not direct plugins, not configured directly) are created by
        classed derived from `AbstractInterpFactory` (which are plugins with configs).
        Each `AbstractInterp` instance will require its own config called `interp_tree_node_config_dict` which are
        cloned/populated by `load_func_envelopes` and indexed by
        the abs path to the interp tree node in this `interp_tree_abs_paths_to_node_configs`.
        """

    def get_plugin_type(
        self,
    ) -> PluginType:
        return PluginType.InterpFactoryPlugin

    def load_interp_tree_abs_paths(
        self,
        this_plugin_instance_interp_tree_abs_paths: list[tuple[str, ...]],
    ):
        """
        This function is meant to be called prior to `load_func_envelopes` to
        let plugins know about all paths in the FS_01_89_09_24 `interp_tree` where they are plugged into.
        """
        pass

    def load_func_envelopes(
        self,
        interp_tree_abs_path: tuple[str, ...],
        func_ids_to_func_envelopes: dict[str, dict],
    ) -> list[str]:
        """
        Load func `data_envelope`-s taking into account `interp_tree_abs_path`.

        Takes part in implementation of FS_01_89_09_24 interp tree.

        Returns list of mapped `func_id`-s.
        """
        if interp_tree_abs_path in self.interp_tree_abs_paths_to_node_configs:
            raise RuntimeError(f"`{interp_tree_abs_path}` has already been loaded")
        else:
            self.interp_tree_abs_paths_to_node_configs[interp_tree_abs_path] = deepcopy(self.plugin_config_dict)

        return []

    def create_interp(
        self,
        interp_ctx: InterpContext,
    ) -> AbstractInterp:
        raise NotImplementedError



from argrelay.relay_client.ClientCommandAbstract import ClientCommandAbstract
from argrelay.server_spec.CallContext import CallContext


class ClientCommandRemoteAbstract(ClientCommandAbstract):

    def __init__(
        self,
        call_ctx: CallContext,
    ):
        super().__init__(
            call_ctx,
        )

    @staticmethod
    def raise_error(
        status_code: int,
    ):
        raise RuntimeError(f"server response status_code: {status_code}")



from __future__ import annotations

from marshmallow import Schema, RAISE, fields

from argrelay.enum_desc.CheckEnvField import CheckEnvField
from argrelay.misc_helper_common.TypeDesc import TypeDesc
from argrelay.schema_config_interp.FuncEnvelopeSchema import (
    func_id_some_func_,
)

field_values_to_command_lines_ = "field_values_to_command_lines"


class SchemaPluginCheckEvnServerResponseValueAbstract(Schema):
    """
    Part of FS_36_17_84_44 `check_env` implementation.
    """

    class Meta:
        unknown = RAISE
        strict = True

    # Configures field name to retrieve in response from command line.
    field_values_to_command_lines = fields.Dict(
        keys = fields.String(),
        values = fields.String(),
        required = True,
    )


schema_plugin_check_evn_server_response_abstract_desc = TypeDesc(
    dict_schema = SchemaPluginCheckEvnServerResponseValueAbstract(),
    ref_name = SchemaPluginCheckEvnServerResponseValueAbstract.__name__,
    dict_example = {
        field_values_to_command_lines_: {
            func_id_some_func_: CheckEnvField.server_git_commit_id.name,
        },
    },
    default_file_path = "",
)



from enum import Enum, auto


class PluginSide(Enum):
    """
    Specifies which side client can be used at.
    """

    PluginServerSideOnly = auto()

    PluginClientSideOnly = auto()

    PluginAnySide = auto()

    def __str__(self):
        return self.name



from __future__ import annotations

from dataclasses import dataclass, field

from argrelay.runtime_data.PluginEntry import PluginEntry


@dataclass
class PluginConfig:
    """
    See also `PluginConfigSchema`.
    """

    reusable_config_data: dict = field()
    """
    Arbitrary data
    """

    plugin_instance_entries: dict[str, PluginEntry] = field()
    """
    Key = `plugin_instance_id`
    """

    plugin_instance_id_activate_list: list[str] = field()
    """
    List of `plugin_instance_id`-s in order of activation generated by
    walking through a DAG described by `plugin_dependencies`.
    Each `plugin_instance_id` is a key into `plugin_instance_entries`.
    """



import signal
from dataclasses import asdict

from argrelay.client_command_remote.ClientCommandRemoteAbstract import ClientCommandRemoteAbstract
from argrelay.client_pipeline.BytesSrcAbstract import BytesSrcAbstract
from argrelay.misc_helper_common import eprint
from argrelay.misc_helper_common.ElapsedTime import ElapsedTime
from argrelay.runtime_data.ConnectionConfig import ConnectionConfig
from argrelay.schema_request.CallContextSchema import call_context_desc
from argrelay.server_spec.CallContext import CallContext
from argrelay.server_spec.const_int import BASE_URL_FORMAT

has_error_happened = False


def _signal_handler(signal_number, signal_frame):
    if signal_number == signal.SIGALRM:
        # Import hanged:
        eprint("ERROR: `import` hanged - see: https://github.com/argrelay/argrelay/issues/89")
        global has_error_happened
        has_error_happened = True


class ClientCommandRemoteWorkerJson(ClientCommandRemoteAbstract):

    def __init__(
        self,
        call_ctx: CallContext,
        connection_config: ConnectionConfig,
        bytes_src: BytesSrcAbstract,
    ):
        super().__init__(
            call_ctx,
        )
        self.connection_config: ConnectionConfig = connection_config
        self.bytes_src: BytesSrcAbstract = bytes_src

    def execute_command(
        self,
    ):
        server_url = BASE_URL_FORMAT.format(**asdict(self.connection_config)) + f"{self.call_ctx.server_action.value}"
        headers_dict = {
            "Content-Type": "application/json",
            "Accept": "application/json",
        }
        # NOTE: So far, single `CallContextSchema` is reused for all client requests:
        request_json = call_context_desc.dict_schema.dumps(self.call_ctx)
        ElapsedTime.measure("before_request")

        # TODO: TODO_30_69_19_14: infinite spinner:
        #     There is a bug - if a child is used, it deadlocks occasionally while importing `requests`:
        #     https://github.com/argrelay/argrelay/issues/89
        #     The workaround is to abort (Ctrl+C) and retry - but this is annoying.
        # Attempt to detect hanging import by setting an alarm at least (not resolving it yet):
        signal.signal(signal.SIGALRM, _signal_handler)
        signal.alarm(1)
        import requests
        signal.alarm(0)

        response_obj = requests.post(
            server_url,
            headers = headers_dict,
            data = request_json,
        )
        ElapsedTime.measure("after_request")
        try:
            if response_obj.ok:
                self.bytes_src.accept_bytes(response_obj.content)
            else:
                self.raise_error(response_obj.status_code)
        finally:
            ElapsedTime.measure("after_handle_response")



from marshmallow import Schema, RAISE, fields, validates_schema, ValidationError

from argrelay.composite_tree.DictTreeWalker import surrogate_node_id_
from argrelay.misc_helper_common.TypeDesc import TypeDesc
from argrelay.plugin_interp.FuncTreeInterpFactory import FuncTreeInterpFactory

interp_selector_tree_ = "interp_selector_tree"

ignored_func_ids_list_ = "ignored_func_ids_list"


def validate_tree_node(interp_selector_sub_tree: dict):
    if isinstance(interp_selector_sub_tree, str):
        return
    elif isinstance(interp_selector_sub_tree, dict):
        for sub_tree_node in interp_selector_sub_tree.values():
            validate_tree_node(sub_tree_node)
    else:
        raise ValidationError(f"neither a str nor a dict: {interp_selector_sub_tree}")


# TODO_40_10_18_32: add custom base to all schemas:
class InterpTreeInterpFactoryConfigSchema(Schema):
    class Meta:
        unknown = RAISE
        strict = True

    interp_selector_tree = fields.Dict(
        keys = fields.String(),
        # TODO: TODO_79_67_28_83: Express recursive dict schema:
        # This is a tree (`dict`) of arbitrary depth with `str` leaves.
        # Ideally, this should be defined as nested `dict`,
        # but it is unknown how to do it in marshmallow.
        values = fields.Raw(),
        required = True,
    )

    # TODO: FS_42_76_93_51: Decide already whether first zero arg interp deserves its future life.
    #                       This `ignored_func_ids_list` field is propagated down from
    #                       the first zero arg interp (needed for its logic as root plugin instance in the composite tree)
    #                       and it also needs to pass validation of this schema (which it converts its config to):
    # TODO: TODO_19_67_22_89: remove `ignored_func_ids_list` - load as `FuncState.fs_unplugged`:
    # It is an error if `func_id` is published by enabled delegator but missing in `func_selector_tree`.
    # To avoid the error, list such `func_id` in `ignored_func_ids_list`.
    ignored_func_ids_list = fields.List(
        fields.String(),
        required = False,
        load_default = [],
    )

    @validates_schema
    def validate_known(
        self,
        input_dict: dict,
        **kwargs,
    ):
        interp_selector_tree = input_dict[interp_selector_tree_]
        if not isinstance(interp_selector_tree, dict):
            raise ValidationError(f"not a dict: {interp_selector_tree}")
        validate_tree_node(interp_selector_tree)


tree_path_interp_factory_config_example = {
    interp_selector_tree_: {
        "some_command": "some_plugin_name",
        surrogate_node_id_: f"{FuncTreeInterpFactory.__name__}.default",
    },
    ignored_func_ids_list_: [
        "func_id_some_ignored",
    ],
}
tree_path_interp_factory_config_desc = TypeDesc(
    dict_schema = InterpTreeInterpFactoryConfigSchema(),
    ref_name = InterpTreeInterpFactoryConfigSchema.__name__,
    dict_example = tree_path_interp_factory_config_example,
    default_file_path = "",
)



from os import getcwd
from unittest import TestCase

from argrelay.misc_helper_common import set_argrelay_dir
from argrelay.test_infra import change_to_known_repo_path


class BaseTestClass(TestCase):
    """
    Root test class with defaults for all tests.
    """

    confusing_result_presence_msg = "remove expected result to avoid confusion"

    @classmethod
    def setUpClass(cls):
        with change_to_known_repo_path("."):
            # Set CWD to the project root = `argrelay_dir` (see FS_29_54_67_86 dir structure):
            set_argrelay_dir(getcwd())
        TestCase.setUpClass()

    @classmethod
    def tearDownClass(cls):
        TestCase.tearDownClass()

    def setUp(self):
        super().setUp()
        self.maxDiff = None

    def tearDown(self):
        super().tearDown()



from __future__ import annotations

from dataclasses import dataclass, field

from argrelay.enum_desc.CompScope import CompScope
from argrelay.enum_desc.ServerAction import ServerAction
from argrelay.enum_desc.TermColor import TermColor
from argrelay.misc_helper_common import eprint


@dataclass(frozen = True)
class CallContext:
    """
    Immutable input data from client to server
    """

    server_action: ServerAction = field()
    command_line: str = field()
    cursor_cpos: int = field()
    comp_scope: CompScope = field()
    client_pid: int = field()
    is_debug_enabled: bool = field()

    def print_debug(
        self,
        end_str: str = "\n",
    ) -> None:
        if not self.is_debug_enabled:
            return
        eprint(TermColor.debug_output.value, end = "")
        eprint(f"server_action: {self.server_action}", end = " ")
        eprint(f"\"{self.command_line}\"", end = " ")
        eprint(f"cursor_cpos: {self.cursor_cpos}", end = " ")
        eprint(f"comp_scope: {self.comp_scope}", end = " ")
        eprint(f"client_pid: {self.client_pid}", end = " ")
        eprint(TermColor.reset_style.value, end = end_str)



from __future__ import annotations

from dataclasses import dataclass, field


@dataclass
class ArgValues:
    """
    See also :class:`ArgValuesSchema`.
    """

    arg_values: list[str] = field()
    """
    Values suggested by `ServerAction.ProposeArgValues`.
    """



import time
import uuid
from copy import deepcopy

from pymongo import MongoClient

from argrelay.composite_tree.DictTreeWalker import contains_whitespace
from argrelay.enum_desc.PluginSide import PluginSide
from argrelay.enum_desc.PluginType import PluginType
from argrelay.enum_desc.ReservedEnvelopeClass import ReservedEnvelopeClass
from argrelay.enum_desc.ReservedPropName import ReservedPropName
from argrelay.enum_desc.SpecialChar import SpecialChar
from argrelay.misc_helper_common import eprint
from argrelay.mongo_data import MongoClientWrapper
from argrelay.mongo_data.MongoClientWrapper import log_validation_progress
from argrelay.mongo_data.MongoServerWrapper import MongoServerWrapper
from argrelay.plugin_delegator.AbstractDelegator import AbstractDelegator
from argrelay.plugin_interp.AbstractInterpFactory import AbstractInterpFactory
from argrelay.plugin_loader.AbstractLoader import AbstractLoader
from argrelay.relay_server.HelpHintCache import HelpHintCache
from argrelay.relay_server.QueryEngine import QueryEngine
from argrelay.runtime_context.AbstractPluginServer import AbstractPluginServer, instantiate_server_plugin
from argrelay.runtime_data.EnvelopeCollection import EnvelopeCollection
from argrelay.runtime_data.PluginConfig import PluginConfig
from argrelay.runtime_data.PluginEntry import PluginEntry
from argrelay.runtime_data.ServerConfig import ServerConfig
from argrelay.runtime_data.StaticData import StaticData
from argrelay.schema_config_core_server.StaticDataSchema import static_data_desc


class LocalServer:
    """
    This is plain server functionality without API-wrapper to expose over the network (hence, local).

    The API-wrapper exposing `LocalServer` over the network is `CustomFlaskApp`.
    """

    def __init__(
        self,
        server_config: ServerConfig,
        plugin_config: PluginConfig,
    ):
        self.server_instance_id = uuid.uuid4()
        self.server_config: ServerConfig = server_config
        self.plugin_config: PluginConfig = plugin_config
        self.mongo_server: MongoServerWrapper = MongoServerWrapper()
        self.mongo_client: MongoClient = MongoClientWrapper.get_mongo_client(self.server_config.mongo_config)
        self.query_engine: QueryEngine = QueryEngine(
            self.server_config.query_cache_config,
            self.get_mongo_database(),
            self.server_config.mongo_config.distinct_values_query,
        )
        self.help_hint_cache: HelpHintCache = HelpHintCache(
            self.query_engine,
        )

        self.cleaned_mongo_collections: set[str] = set()
        """
        *   On one hand, Mongo collections may have state and need to be cleaned.
        *   On another hand, they can repeatedly loaded by more than one plugin (potentially adding more data).

        Server uses `cleaned_mongo_collections` to track what collection have been cleaned once and what have not.
        """

        # TODO: TODO_39_25_11_76: missing props: index to validate missing props:
        # Set of all prop names (2-level map) per `collection_name` and `envelope_class`:
        self.prop_names_per_collection_per_class: dict[str, dict[str, set[str]]] = {}

        # Number of envelopes loaded per `collection_name`:
        self.count_per_collection: dict[str, int] = {}

        # seconds since epoch:
        self.server_start_time: int = int(time.time())

    def start_local_server(
        self,
    ):
        self._activate_plugins()
        self._start_mongo_server()
        self._load_mongo_data()
        self._create_mongo_index()
        self._populate_help_hint_cache()
        self._log_connection_url()

    def stop_local_server(
        self,
    ):
        self._stop_mongo_server()

    def get_mongo_database(
        self,
    ):
        return self.mongo_client[self.server_config.mongo_config.mongo_server.database_name]

    def get_query_engine(
        self,
    ):
        return self.query_engine

    def _activate_plugins(
        self,
    ):
        """
        Calls each plugin to update :class:`StaticData`.
        """

        for plugin_instance_id in self.plugin_config.plugin_instance_id_activate_list:
            plugin_entry: PluginEntry = self.plugin_config.plugin_instance_entries[plugin_instance_id]

            if not plugin_entry.plugin_enabled:
                continue

            if (
                plugin_entry.plugin_side != PluginSide.PluginServerSideOnly
                and
                plugin_entry.plugin_side != PluginSide.PluginAnySide
            ):
                continue

            plugin_instance: AbstractPluginServer = instantiate_server_plugin(
                self.server_config,
                plugin_instance_id,
                plugin_entry,
            )
            plugin_type = plugin_instance.get_plugin_type()

            assert plugin_type.plugin_side is plugin_entry.plugin_side

            if plugin_type is PluginType.LoaderPlugin:
                plugin_instance: AbstractLoader
                plugin_instance.activate_plugin()
                # Store instance of `AbstractLoader` under specified id for future use:
                self.server_config.data_loaders[plugin_instance_id] = plugin_instance
                continue

            if plugin_type is PluginType.InterpFactoryPlugin:
                plugin_instance: AbstractInterpFactory
                plugin_instance.activate_plugin()
                # Store instance of `AbstractInterpFactory` under specified id for future use:
                self.server_config.interp_factories[plugin_instance_id] = plugin_instance
                continue

            if plugin_type is PluginType.DelegatorPlugin:
                plugin_instance: AbstractDelegator
                plugin_instance.activate_plugin()
                # Store instance of `AbstractDelegator` under specified id for future use:
                self.server_config.action_delegators[plugin_instance_id] = plugin_instance
                continue

            if plugin_type is PluginType.ConfiguratorPlugin:
                plugin_instance: AbstractDelegator
                plugin_instance.activate_plugin()
                # Store instance of `AbstractConfigurator` under specified id for future use:
                self.server_config.server_configurators[plugin_instance_id] = plugin_instance
                continue

    def _pre_validate_static_data(
        self,
    ):
        self._pre_validate_static_data_schema()
        self._pre_validata_static_data_by_plugins()
        self._pre_validate_data_envelope_string_prop_values()

    def _pre_validate_static_data_schema(
        self,
    ):
        # Note that this is slow for large data sets:
        static_data_dict = static_data_desc.dict_schema.dump(self.server_config.static_data)
        static_data_desc.validate_dict(static_data_dict)

    def _pre_validata_static_data_by_plugins(
        self,
    ):
        all_plugins: [AbstractLoader] = []
        all_plugins.extend(self.server_config.data_loaders.values())
        all_plugins.extend(self.server_config.interp_factories.values())
        all_plugins.extend(self.server_config.action_delegators.values())

        for plugin_instance in all_plugins:
            plugin_instance.validate_loaded_data(self.server_config.static_data)

    def _pre_validate_data_envelope_string_prop_values(
        self,
    ):
        """
        This validation ensures there is no blank values (`None`, whitespace, etc.) in `index_prop`-s.

        See also FS_99_81_19_25 no space in options.
        """

        for collection_name, envelope_collection in self.server_config.static_data.envelope_collections.items():
            prop_names_per_class: dict[str, set[str]] = self.prop_names_per_collection_per_class.setdefault(
                collection_name,
                {},
            )
            self.count_per_collection[collection_name] = (
                self.count_per_collection.setdefault(collection_name, 0)
                +
                len(envelope_collection.data_envelopes)
            )
            for index_prop in envelope_collection.index_props:
                for data_envelope in envelope_collection.data_envelopes:
                    envelope_class = data_envelope[ReservedPropName.envelope_class.name]

                    if index_prop not in data_envelope:
                        # TODO_39_25_11_76: `data_envelope`-s with missing props:
                        # Let the `data_envelope` load without some `prop_name`-s from `index_prop`-s -
                        # if any of `data_envelope`-s have that `prop_name`, it will fail validation.
                        # It is allowed to have no `prop_name` for all `data_envelope` until one has it.
                        continue
                    else:
                        prop_names_per_class.setdefault(envelope_class, set()).add(index_prop)

                    prop_value = data_envelope[index_prop]
                    self._validate_string_prop_value(
                        envelope_class,
                        index_prop,
                        prop_value,
                    )

    def _validate_string_prop_value(
        self,
        envelope_class,
        index_prop,
        prop_value,
    ):
        if isinstance(prop_value, str):
            if not prop_value and prop_value.strip():
                raise ValueError(f"`{envelope_class}.{index_prop}` [{prop_value}] has to be non-blank string")
            if contains_whitespace(prop_value):
                raise ValueError(f"`{envelope_class}.{index_prop}` [{prop_value}] cannot contain whitespace")
        elif isinstance(prop_value, list):
            for prop_value_item in prop_value:
                self._validate_string_prop_value(
                    envelope_class,
                    index_prop,
                    prop_value_item,
                )
        else:
            # FS_06_99_43_60: list arg value:
            raise ValueError(f"`{envelope_class}.{index_prop}` has to be a `list` of `str` or `str`")

    def _post_validate_loaded_data(
        self,
    ):
        self._post_validate_data_envelope_missing_prop_names()

    def _post_validate_data_envelope_missing_prop_names(
        self,
    ):
        """
        See TODO_39_25_11_76 missing props.
        """

        # Verify that all prop names per (2-level map) `collection_name` and `envelope_class`
        # exists in all corresponding `data_envelope`-s loaded:
        for collection_name, prop_names_per_class in self.prop_names_per_collection_per_class.items():
            eprint(f"collection to validate: {collection_name}")
            total_envelope_n = self.count_per_collection[collection_name]
            curr_envelope_i = 0
            log_validation_progress(collection_name, curr_envelope_i, total_envelope_n)

            for data_envelope in self.query_engine.get_data_envelopes_cursor(
                collection_name,
                {},
            ):
                curr_envelope_i += 1
                envelope_class = data_envelope[ReservedPropName.envelope_class.name]

                if curr_envelope_i % 1_000 == 0:
                    log_validation_progress(collection_name, curr_envelope_i, total_envelope_n)

                for prop_name in prop_names_per_class[envelope_class]:
                    if prop_name not in data_envelope:
                        raise ValueError(
                            f"data_envelope of (collection_name: `{collection_name}`, envelope_class: `{envelope_class}`) does not have (prop_name: `{prop_name}`): {data_envelope}"
                        )
                    prop_value = data_envelope[prop_name]
                    self._validate_string_prop_value(
                        envelope_class,
                        prop_name,
                        prop_value,
                    )

            log_validation_progress(collection_name, curr_envelope_i, total_envelope_n)

    def _start_mongo_server(
        self,
    ):
        self.mongo_server.start_mongo_server(self.server_config.mongo_config)

    def _stop_mongo_server(
        self,
    ):
        self.mongo_server.stop_mongo_server()

    def _populate_func_missing_props(
        self,
    ):
        """
        TODO: TODO_39_25_11_76 missing props: populate missing props supported by funcs.

        This function only targets `ReservedEnvelopeClass.ClassFunction` (because loading funcs is special).
        There is no decision (yet) to populate missing `prop_value`-s for any `data_envelope`.
        Instead, there is a validation to prevent missing `prop_name`-s - each loader has to be fixed
        to provide seme set of `prop_name`-s for all `data_envelope`.
        """

        # Collect `prop_name`-s used by all funcs:
        func_prop_names: set[str] = set()
        for mongo_collection in self.server_config.static_data.envelope_collections:
            envelope_collection: EnvelopeCollection = self.server_config.static_data.envelope_collections[
                mongo_collection
            ]
            for data_envelope in envelope_collection.data_envelopes:
                envelope_class = data_envelope[ReservedPropName.envelope_class.name]
                if envelope_class == ReservedEnvelopeClass.ClassFunction.name:
                    for prop_name in envelope_collection.index_props:
                        func_prop_names.add(prop_name)

        # Populate missing `prop_name`-s with special value:
        for mongo_collection in self.server_config.static_data.envelope_collections:
            envelope_collection: EnvelopeCollection = self.server_config.static_data.envelope_collections[
                mongo_collection
            ]
            for data_envelope in envelope_collection.data_envelopes:
                envelope_class = data_envelope[ReservedPropName.envelope_class.name]
                if envelope_class == ReservedEnvelopeClass.ClassFunction.name:
                    for prop_name in func_prop_names:
                        if prop_name not in data_envelope:
                            data_envelope[prop_name] = SpecialChar.NoPropValue.value

    def _load_mongo_data(
        self,
    ):

        total_envelope_n: int = 0
        curr_envelope_i: int = 0

        # At this moment, funcs have already been loaded on `AbstractPlugin.activate_plugin`.
        self._populate_func_missing_props()

        # TODO_00_79_72_55: Remove `static_data` from `server_config`:
        # Initial step: load funcs and any data from config:
        self._load_mongo_data_step(
            "config_data",
            total_envelope_n,
            curr_envelope_i,
        )

        for plugin_instance_id in self.plugin_config.plugin_instance_id_activate_list:
            if plugin_instance_id in self.server_config.data_loaders:
                plugin_instance = self.server_config.data_loaders[plugin_instance_id]
                self.server_config.static_data = StaticData(
                    envelope_collections = {},
                )
                # Use loader to update data:
                self.server_config.static_data = plugin_instance.update_static_data(
                    self.server_config.static_data,
                    self.query_engine,
                )
                self._load_mongo_data_step(
                    plugin_instance_id,
                    total_envelope_n,
                    curr_envelope_i,
                )

        self._post_validate_loaded_data()

    def _load_mongo_data_step(
        self,
        step_name: str,
        total_envelope_n: int,
        curr_envelope_i: int,
    ):
        mongo_db = self.mongo_client[self.server_config.mongo_config.mongo_server.database_name]

        eprint(f"{step_name}:")

        self._pre_validate_static_data()

        MongoClientWrapper.store_envelopes(
            mongo_db,
            self.cleaned_mongo_collections,
            # TODO_00_79_72_55: Remove `static_data` from `server_config`:
            self.server_config.static_data,
            total_envelope_n,
            curr_envelope_i,
        )

    def _create_mongo_index(
        self,
    ):
        mongo_db = self.mongo_client[self.server_config.mongo_config.mongo_server.database_name]

        for collection_name in self.server_config.static_data.envelope_collections:
            envelope_collection: EnvelopeCollection = self.server_config.static_data.envelope_collections[
                collection_name
            ]
            # Include `envelope_class` field into index by default:
            index_props: list[str] = deepcopy(envelope_collection.index_props)
            index_props.append(ReservedPropName.envelope_class.name)
            MongoClientWrapper.create_index(
                mongo_db,
                collection_name,
                envelope_collection.index_props,
            )

    def _populate_help_hint_cache(
        self,
    ):
        self.help_hint_cache.populate_cache()

    def _log_connection_url(
        self,
    ):
        host_name = self.server_config.connection_config.server_host_name
        port_number = self.server_config.connection_config.server_port_number
        eprint(f"http://{host_name}:{port_number}")



"""
"""
from __future__ import annotations

from typing import Callable, Any, Union

from argrelay.enum_desc.ArgSource import ArgSource
from argrelay.plugin_config.AbstractConfigurator import AbstractConfigurator
from argrelay.runtime_data.AssignedValue import AssignedValue


def set_default_to(
    prop_name,
    prop_value,
    envelope_container,
) -> bool:
    """
    Assign a default `prop_value` to the given `prop_name`:
    *   Ensure that this `prop_name` is valid for `envelope_container`.
    *   Ensure that this `prop_value` is actually one of the remaining ones.
    """

    if prop_name in envelope_container.search_control.types_to_keys_dict:
        if prop_name not in envelope_container.assigned_types_to_values:
            if prop_name in envelope_container.remaining_types_to_values:
                if prop_value in envelope_container.remaining_types_to_values[prop_name]:
                    del envelope_container.remaining_types_to_values[prop_name]
                    envelope_container.assigned_types_to_values[prop_name] = AssignedValue(
                        prop_value,
                        ArgSource.DefaultValue,
                    )
                    return True
        else:
            # Setting `ArgSource.DefaultValue`: it cannot override any, right? No point to handle assigned case:
            pass
    return False


def get_config_value_once(
    server_configurators,
    value_getter: Callable[[AbstractConfigurator], Any],
    default_value: Any,
) -> Any:
    config_value: Union[Any, None] = None
    server_configurator: AbstractConfigurator
    for server_configurator in server_configurators.values():
        if config_value is None:
            config_value = value_getter(server_configurator)
        else:
            if value_getter(server_configurator) is not None:
                # Only one `PluginType.ConfiguratorPlugin` providing
                # same type of value is supported to avoid confusion:
                raise RuntimeError

    if config_value is None:
        config_value = default_value
    return config_value



from argrelay.check_env.CheckEnvResult import CheckEnvResult
from argrelay.check_env.PluginCheckEnvServerResponseValueAbstract import PluginCheckEnvServerResponseValueAbstract
from argrelay.custom_integ.SchemaPluginCheckEvnServerResponseValueAbstract import field_values_to_command_lines_
from argrelay.custom_integ.git_utils import is_git_repo, get_full_commit_id
from argrelay.enum_desc.CheckEnvField import CheckEnvField
from argrelay.enum_desc.ResultCategory import ResultCategory
from argrelay.misc_helper_common import get_argrelay_dir


class PluginCheckEnvServerResponseValueCommitId(PluginCheckEnvServerResponseValueAbstract):

    def __init__(
        self,
        plugin_instance_id: str,
        plugin_config_dict: dict,
    ):
        super().__init__(
            plugin_instance_id,
            plugin_config_dict = plugin_config_dict or {
                # TODO: Maybe remove this class and simply rely on config?
                field_values_to_command_lines_: {
                    CheckEnvField.server_git_commit_id.name: "argrelay.check_env server_commit",
                },
            },
        )

    def verify_value(
        self,
        field_name,
        field_value,
    ) -> CheckEnvResult:
        if field_name == CheckEnvField.server_git_commit_id.name:

            argrelay_dir = get_argrelay_dir()

            if is_git_repo(argrelay_dir):
                try:
                    client_commit_id = get_full_commit_id(argrelay_dir)
                except ValueError:
                    client_commit_id = None
            else:
                client_commit_id = None

            if client_commit_id is None:
                return CheckEnvResult(
                    result_category = ResultCategory.ExecutionFailure,
                    result_key = field_name,
                    result_value = field_value,
                    result_message = "failed on getting client commit id",
                )
            else:
                if client_commit_id == field_value:
                    return CheckEnvResult(
                        result_category = ResultCategory.VerificationSuccess,
                        result_key = field_name,
                        result_value = field_value,
                        result_message = f"client commit id [{client_commit_id}] matches server commit id",
                    )
                else:
                    return CheckEnvResult(
                        result_category = ResultCategory.VerificationWarning,
                        result_key = field_name,
                        result_value = field_value,
                        result_message = f"client commit id [{client_commit_id}] does not match server commit id [{field_value}]",
                    )
        else:
            return super().verify_value(
                field_name,
                field_value,
            )



from marshmallow import RAISE, fields

from argrelay.misc_helper_common.ObjectSchema import ObjectSchema
from argrelay.misc_helper_common.TypeDesc import TypeDesc
from argrelay.runtime_context.EnvelopeContainer import EnvelopeContainer
from argrelay.schema_config_interp.DataEnvelopeSchema import data_envelope_desc
from argrelay.schema_config_interp.SearchControlSchema import (
    search_control_desc,
)
from argrelay.schema_response.AssignedValueSchema import assigned_value_desc

"""
Schema for :class:`EnvelopeContainer`
"""

search_control_ = "search_control"
data_envelopes_ = "data_envelopes"
found_count_ = "found_count"
used_arg_bucket_ = "used_arg_bucket"
assigned_types_to_values_ = "assigned_types_to_values"
remaining_types_to_values_ = "remaining_types_to_values"
filled_types_to_values_hidden_by_defaults_ = "filled_types_to_values_hidden_by_defaults"


class EnvelopeContainerSchema(ObjectSchema):
    class Meta:
        unknown = RAISE
        ordered = True

    model_class = EnvelopeContainer

    search_control = fields.Nested(
        search_control_desc.dict_schema,
        required = True,
    )

    data_envelopes = fields.List(
        fields.Nested(data_envelope_desc.dict_schema),
        required = True,
    )

    found_count = fields.Integer(
        required = True,
    )

    used_arg_bucket = fields.Integer(
        required = True,
        allow_none = True,
    )

    assigned_types_to_values = fields.Dict(
        keys = fields.String(),
        values = fields.Nested(assigned_value_desc.dict_schema),
        required = True,
    )

    remaining_types_to_values = fields.Dict(
        keys = fields.String(),
        values = fields.List(
            fields.String(),
        ),
        required = True,
    )

    filled_types_to_values_hidden_by_defaults = fields.Dict(
        keys = fields.String(),
        values = fields.List(
            fields.String(),
        ),
        required = True,
    )


envelope_container_desc = TypeDesc(
    dict_schema = EnvelopeContainerSchema(),
    ref_name = EnvelopeContainerSchema.__name__,
    dict_example = {
        search_control_: search_control_desc.dict_example,
        data_envelopes_: [
            data_envelope_desc.dict_example,
        ],
        found_count_: 1,
        used_arg_bucket_: 0,
        assigned_types_to_values_: {
            "SomeTypeA": assigned_value_desc.dict_example,
        },
        remaining_types_to_values_: {
            "SomeTypeB": [
                "B_value_1",
                "B_value_7",
            ],
            "SomeTypeC": [
                "C_value_3",
                "C_value_9",
            ],
        },
        filled_types_to_values_hidden_by_defaults_: {
        },
    },
    default_file_path = "",
)



from dataclasses import dataclass, field


@dataclass
class QueryCacheConfig:
    enable_query_cache: bool = field(default = True)

    query_cache_ttl_sec: int = field(default = 60)

    query_cache_max_size_bytes: int = field(default = 1024 * 1024)



from __future__ import annotations

from copy import deepcopy

from marshmallow import RAISE, Schema, fields, validates_schema, pre_load

from argrelay.custom_integ.BaseConfigDelegatorConfigSchema import BaseConfigDelegatorConfigSchema, func_configs_
from argrelay.custom_integ.FuncConfigSchema import func_config_desc, func_envelope_
from argrelay.misc_helper_common.TypeDesc import TypeDesc
from argrelay.schema_config_interp.DataEnvelopeSchema import envelope_payload_
from argrelay.schema_config_interp.FuncEnvelopeSchema import (
    func_id_some_func_,
)

command_template_ = "command_template"
echo_command_on_stderr_ = "echo_command_on_stderr"


class ConfigOnlyDelegatorConfigSchema(BaseConfigDelegatorConfigSchema):
    """
    Part of FS_49_96_50_77 config_only_delegator implementation.
    """

    class Meta:
        unknown = RAISE
        strict = True

    @pre_load
    def pre_load(
        self,
        input_dict: dict,
        **kwargs
    ) -> dict:
        """
        See `validate_data` why this has to be done:
        """
        for func_config in input_dict[func_configs_].values():
            func_config[func_envelope_][
                envelope_payload_
            ] = config_only_delegator_envelope_payload_desc.dict_from_input_dict(
                func_config[func_envelope_][envelope_payload_]
            )
        return input_dict

    @validates_schema
    def validate_data(
        self,
        input_dict: dict,
        **kwargs,
    ):
        """
        `BaseConfigDelegatorConfigSchema.func_configs` dict contains values with
        `FuncConfigSchema.func_envelope` of `FuncEnvelopeSchema` which allows arbitrary dict in `envelope_payload`.

        `ConfigOnlyDelegatorConfigSchema` expects them to be of `ConfigOnlyDelegatorEnvelopePayloadSchema`.
        """
        for func_config in input_dict[func_configs_].values():
            config_only_delegator_envelope_payload_desc.validate_dict(func_config[func_envelope_][envelope_payload_])


class ConfigOnlyDelegatorEnvelopePayloadSchema(Schema):
    class Meta:
        unknown = RAISE
        strict = True

    command_template = fields.String(
        required = True,
    )

    echo_command_on_stderr = fields.Boolean(
        required = False,
        load_default = False,
    )

    # TODO_54_68_18_12: Support defaults for config-only delegator:
    # Add a schema for this config but likely into _new_ `BaseConfigDelegatorEnvelopePayloadSchema`
    # as it can be used with other configurable plugins even if they do not use
    # `command_template` (as this `ConfigOnlyDelegatorEnvelopePayloadSchema` class does).

    # TODO_74_73_60_93: Support expected envelope count in config-only delegator:
    # Similar to TODO_54_68_18_12 (above), it should be part of _new_ `BaseConfigDelegatorEnvelopePayloadSchema`.


_config_only_func_config_dict_example = deepcopy(func_config_desc.dict_example)
_config_only_func_config_dict_example[func_envelope_][envelope_payload_].update({
    command_template_: "echo",
})

config_only_delegator_envelope_payload_desc = TypeDesc(
    dict_schema = ConfigOnlyDelegatorEnvelopePayloadSchema(),
    ref_name = ConfigOnlyDelegatorEnvelopePayloadSchema.__name__,
    dict_example = _config_only_func_config_dict_example,
    default_file_path = "",
)

config_only_delegator_config_desc = TypeDesc(
    dict_schema = ConfigOnlyDelegatorConfigSchema(),
    ref_name = ConfigOnlyDelegatorConfigSchema.__name__,
    dict_example = {
        func_configs_: {
            func_id_some_func_: config_only_delegator_envelope_payload_desc.dict_example,
        },
    },
    default_file_path = "",
)



from __future__ import annotations

import traceback
from typing import cast

from argrelay.check_env.CheckEnvResult import CheckEnvResult
from argrelay.check_env.PluginCheckEnvAbstract import PluginCheckEnvAbstract
from argrelay.client_command_remote.ClientCommandRemoteWorkerJson import ClientCommandRemoteWorkerJson
from argrelay.client_pipeline.BytesHandlerJson import BytesHandlerJson
from argrelay.client_pipeline.BytesSrcLocal import BytesSrcLocal
from argrelay.client_spec.ShellContext import ShellContext, UNKNOWN_COMP_KEY
from argrelay.custom_integ.SchemaPluginCheckEvnServerResponseValueAbstract import (
    field_values_to_command_lines_,
    schema_plugin_check_evn_server_response_abstract_desc,
)
from argrelay.enum_desc.CompType import CompType
from argrelay.enum_desc.ProcRole import ProcRole
from argrelay.enum_desc.ResultCategory import ResultCategory
from argrelay.handler_response.ClientResponseHandlerCheckEnv import ClientResponseHandlerCheckEnv
from argrelay.misc_helper_common import eprint
from argrelay.relay_client.proc_worker import worker_main
from argrelay.runtime_data.ClientConfig import ClientConfig
from argrelay.schema_config_core_client.ClientConfigSchema import client_config_desc
from argrelay.schema_response.InvocationInput import InvocationInput


class PluginCheckEnvServerResponseValueAbstract(PluginCheckEnvAbstract):

    def __init__(
        self,
        plugin_instance_id: str,
        plugin_config_dict: dict,
    ):
        self.response_value_config_desc = schema_plugin_check_evn_server_response_abstract_desc
        super().__init__(
            plugin_instance_id,
            plugin_config_dict,
        )

    def validate_config(
        self,
    ) -> None:
        self.response_value_config_desc.validate_dict(self.plugin_config_dict)

    def execute_check(
        self,
    ) -> list[CheckEnvResult]:

        check_env_results: list[CheckEnvResult] = []
        for field_name, command_line in self.plugin_config_dict[field_values_to_command_lines_].items():
            # noinspection PyBroadException
            try:
                shell_ctx = ShellContext(
                    command_line = command_line,
                    cursor_cpos = len(command_line),
                    comp_type = CompType.InvokeAction,
                    comp_key = UNKNOWN_COMP_KEY,
                    is_debug_enabled = False,
                )
                client_config: ClientConfig = client_config_desc.obj_from_default_file()

                call_ctx = shell_ctx.create_call_context()

                proc_role: ProcRole = ProcRole.CheckEnvWorker
                command_obj: ClientCommandRemoteWorkerJson = worker_main(
                    call_ctx,
                    client_config,
                    proc_role,
                    False,
                    None,
                    shell_ctx,
                )

                invocation_input: InvocationInput = cast(
                    ClientResponseHandlerCheckEnv,
                    cast(
                        BytesHandlerJson,
                        cast(
                            BytesSrcLocal,
                            command_obj.bytes_src,
                        ).bytes_handler,
                    ).client_response_handler,
                ).invocation_input

                check_env_results.append(self.verify_value(
                    field_name,
                    invocation_input.custom_plugin_data[field_name],
                ))

            except Exception as e:
                eprint(traceback.format_exc())
                check_env_results.append(CheckEnvResult(
                    result_category = ResultCategory.ExecutionFailure,
                    result_key = field_name,
                    result_value = None,
                    result_message = f"{str(type(e))}: {str(e)}",
                ))
        return check_env_results

    def verify_value(
        self,
        field_name,
        field_value,
    ) -> CheckEnvResult:
        """
        Default implementation verifies nothing.
        """
        return CheckEnvResult(
            result_category = ResultCategory.VerificationSuccess,
            result_key = field_name,
            result_value = field_value,
            result_message = None,
        )



from __future__ import annotations

from marshmallow import RAISE, fields

from argrelay.enum_desc.ReservedEnvelopeClass import ReservedEnvelopeClass
from argrelay.misc_helper_common.ObjectSchema import ObjectSchema
from argrelay.misc_helper_common.TypeDesc import TypeDesc
from argrelay.runtime_data.StaticData import StaticData
from argrelay.schema_config_core_server.EnvelopeCollectionSchema import envelope_collection_desc

envelope_collections_ = "envelope_collections"


# TODO_00_79_72_55: remove in the future:
class StaticDataSchema(ObjectSchema):
    class Meta:
        unknown = RAISE
        strict = True

    model_class = StaticData

    envelope_collections = fields.Dict(
        keys = fields.String(),
        values = fields.Nested(envelope_collection_desc.dict_schema),
        required = True,
    )


static_data_desc = TypeDesc(
    dict_schema = StaticDataSchema(),
    ref_name = StaticDataSchema.__name__,
    dict_example = {
        envelope_collections_: {
            ReservedEnvelopeClass.ClassFunction.name: envelope_collection_desc.dict_example,
        },
    },
    default_file_path = "",
)



from enum import Enum


class TermColor(Enum):
    """
    Color codes for terminal text

    Reference:
    *   https://pkg.go.dev/github.com/whitedevops/colors
    *   https://gist.github.com/vratiu/9780109
    """

    ###################################################################################################################
    # Direct colors:
    # do not use them directly, use semantic colors instead (below).

    back_dark_red = "\033[41m"
    back_dark_green = "\033[42m"
    back_dark_yellow = "\033[43m"
    back_dark_blue = "\033[44m"

    back_bright_red = "\033[101m"
    back_bright_green = "\033[102m"
    back_bright_yellow = "\033[103m"

    fore_dark_red = "\033[31m"
    fore_dark_green = "\033[32m"
    fore_dark_yellow = "\033[33m"
    fore_dark_blue = "\033[34m"
    fore_dark_magenta = "\033[35m"
    fore_dark_cyan = "\033[36m"
    fore_dark_gray = "\033[90m"

    fore_bright_gray = "\033[90m"
    fore_bright_red = "\033[91m"
    fore_bright_green = "\033[92m"
    fore_bright_yellow = "\033[93m"
    fore_bright_blue = "\033[94m"
    fore_bright_magenta = "\033[95m"
    fore_bright_cyan = "\033[96m"
    fore_bright_white = "\033[97m"

    fore_bold_dark_red = "\033[1;31m"

    ###################################################################################################################
    # Semantic colors:

    prefix_highlight = back_dark_blue

    known_envelope_id = fore_dark_gray
    unknown_envelope_id = fore_dark_red

    no_help_hint = fore_dark_red

    help_hint = fore_dark_green

    tangent_token_l_part = fore_bright_white
    """
    See `ParsedContext.tan_token_l_part`
    """

    tangent_token_r_part = fore_bright_cyan
    """
    See `ParsedContext.tangent_token_r_part`
    """

    explicit_pos_arg_value = fore_bright_blue
    """
    See `ArgSource.ExplicitPosArg`.
    """

    other_assigned_arg_value = fore_dark_green
    """
    Any other assigned arg value except `explicit_pos_arg_value`.
    """

    remaining_value = fore_bright_yellow
    """
    See `EnvelopeContainer.remaining_types_to_values`.
    """

    excluded_token = fore_dark_blue
    """
    See:
    *   `InterpContext.excluded_tokens`
    *   `InterpResult.excluded_tokens`
    """

    consumed_token = fore_bright_blue
    """
    See:
    *   `InterpContext.consumed_arg_buckets`
    *   `InterpResult.consumed_arg_buckets`
    """

    remaining_token = fore_bright_magenta
    """
    See:
    *   `InterpContext.remaining_arg_buckets`
    """

    no_option_to_suggest = fore_dark_gray
    """
    When none of the relevant `data_envelope`-s has values for the given prop.
    """

    caption_hidden_by_default = fore_dark_magenta
    """
    FS_72_53_55_13: Caption for values hidden by default.
    """

    value_hidden_by_default = fore_dark_yellow
    """
    FS_72_53_55_13: Option hidden by default.
    """

    found_count_0 = back_dark_red
    found_count_1 = fore_dark_green
    found_count_n = fore_bright_yellow

    spinner_color = fore_bold_dark_red

    debug_output = fore_dark_gray

    reset_style = "\033[0m"



from __future__ import annotations

from dataclasses import dataclass, field


@dataclass(frozen = True)
class EnvelopeCollection:
    """
    Allows distributing `data_envelope`-s across FS_56_43_05_79 different collections.

    Specifies collection of `data_envelope`-s to load and associated `index_props` to create.

    See also `EnvelopeCollectionSchema`.
    """

    index_props: list = field()
    """
    Lists fields of `data_envelop` which are to be indexed by MongoDB for given collection.
    """

    data_envelopes: list = field()



from __future__ import annotations

from enum import Enum, auto
from typing import Union

from argrelay.composite_tree.AbstractNodeVisitor import AbstractNodeVisitor
from argrelay.composite_tree.CompositeForest import CompositeForest
from argrelay.composite_tree.CompositeInfoType import CompositeInfoType
from argrelay.composite_tree.CompositeNode import (
    InterpTreeNode,
    TreePathNode,
    FuncTreeNode,
    ZeroArgNode,
    BaseNode,
)
from argrelay.composite_tree.DictTreeWalker import surrogate_node_id_


class TraverseDecision(Enum):
    walk_sub_tree = auto()
    skip_sub_tree = auto()


def _put_value_into_dict_path(
    given_value,
    given_path: list[str],
    target_dict: dict,
) -> None:
    curr_sub_dict: dict = target_dict

    assert len(given_path) > 0

    if len(given_path) > 1:
        for path_step in given_path[:-1]:
            if path_step not in curr_sub_dict:
                # create:
                curr_sub_dict[path_step] = {}
            curr_sub_dict = curr_sub_dict[path_step]

    curr_sub_dict[given_path[-1]] = given_value


class AbstractCompositeTreeWalker(AbstractNodeVisitor):
    """
    FS_33_76_82_84 composite tree walker

    It implements traversal algorithms which extract various `CompositeInfoType`-s
    (info in a form simplified for specific purpose).

    The original intention is to extract info in the form which `DictTreeWalker` can consume -
    an intermediate phase to migrate to composite tree config while still using existing code
    (existing code relied on data structures processed via `DictTreeWalker`).
    """

    def __init__(
        self,
        composite_forest: CompositeForest,
        info_type: CompositeInfoType,
    ):
        # main input:
        self.tree_roots: dict[str, ZeroArgNode] = composite_forest.tree_roots
        self.info_type: CompositeInfoType = info_type
        self.curr_node_id: Union[str, None] = None
        self.curr_node: Union[ZeroArgNode, InterpTreeNode, FuncTreeNode, TreePathNode, None] = None
        self.curr_path: list[str] = []

        # internal trackers:
        self.visitor_decision: TraverseDecision = TraverseDecision.walk_sub_tree

    def walk_tree_roots(
        self,
    ) -> None:
        for tree_root_id in self.tree_roots.keys():
            self._descend_and_ascend(
                self.tree_roots,
                tree_root_id,
            )

    def _walk_tree_node(
        self,
    ) -> None:
        traverse_decision = self._process_tree_node()
        if traverse_decision is TraverseDecision.walk_sub_tree:
            self._walk_sub_tree()
        elif traverse_decision is TraverseDecision.skip_sub_tree:
            pass

    def _walk_sub_tree(
        self,
    ):
        if self.curr_node.sub_tree is not None:
            for sub_tree_node_id in self.curr_node.sub_tree.keys():
                self._descend_and_ascend(
                    self.curr_node.sub_tree,
                    sub_tree_node_id,
                )

    def _descend_and_ascend(
        self,
        node_dict: dict[str, BaseNode],
        node_id: str,
    ) -> None:
        """
        Recursive DFS traversal of `composite_tree`.
        """
        # prepare:
        self.curr_path.append(node_id)
        prev_node = self.curr_node
        prev_node_id = self.curr_node_id
        self.curr_node_id = node_id
        # dive:
        self.curr_node = node_dict[node_id]
        self._walk_tree_node()
        # recover:
        del self.curr_path[-1]
        self.curr_node = prev_node
        self.curr_node_id = prev_node_id

    def _process_tree_node(
        self,
    ) -> TraverseDecision:
        self.visit_node(self.curr_node)
        return self.visitor_decision


# noinspection PyPep8Naming
class _CompositeTreeWalker_tree_abs_path_to_interp_id(AbstractCompositeTreeWalker):

    def __init__(
        self,
        composite_forest: CompositeForest,
        info_type: CompositeInfoType,
        func_id: Union[str, None],
    ):
        # main input:
        super().__init__(
            composite_forest,
            info_type,
        )

        # `info_type`-specific input:
        self.func_id: Union[str, None] = func_id

        # internal trackers:
        self.last_next_interp: Union[InterpTreeNode.NextInterp, None] = None

        # results:
        self.tree_abs_path_to_interp_id: dict = {}

    # noinspection PyPep8Naming
    def _visit_ZeroArgNode(self, node: ZeroArgNode) -> None:
        self.visitor_decision = TraverseDecision.walk_sub_tree

    # noinspection PyPep8Naming
    def _visit_TreePathNode(self, node: TreePathNode) -> None:
        self.visitor_decision = TraverseDecision.walk_sub_tree

    # noinspection PyPep8Naming
    def _visit_InterpTreeNode(self, node: InterpTreeNode) -> None:
        self.last_next_interp = self.curr_node.next_interp
        self.visitor_decision = TraverseDecision.walk_sub_tree

    # noinspection PyPep8Naming
    def _visit_FuncTreeNode(self, node: FuncTreeNode) -> None:
        if self.curr_node.func_id == self.func_id:
            assert self.last_next_interp is not None
            assert self.curr_path[-1] == surrogate_node_id_
            assert self.curr_node.sub_tree is None
            _put_value_into_dict_path(
                self.last_next_interp.plugin_instance_id,
                self.curr_path[:-1],
                self.tree_abs_path_to_interp_id,
            )
        self.visitor_decision = TraverseDecision.skip_sub_tree


# noinspection PyPep8Naming
class _CompositeTreeWalker_zero_arg_interp_tree(AbstractCompositeTreeWalker):

    def __init__(
        self,
        composite_forest: CompositeForest,
        info_type: CompositeInfoType,
    ):
        # main input:
        super().__init__(
            composite_forest,
            info_type,
        )

        # results:
        self.zero_arg_interp_tree: dict = {}

    # noinspection PyPep8Naming
    def _visit_ZeroArgNode(self, node: ZeroArgNode) -> None:
        assert len(self.curr_path) == 1
        _put_value_into_dict_path(
            # TODO: TODO_18_51_46_14: refactor FS_42_76_93_51 zero_arg_interp into FS_15_79_76_85 line processor:
            #       There could be multiple `zero_arg_node`-s each with its own interp,
            #       but now, there is only one interp selected by `first_interp_factory_id`.
            #       In fact, `first_interp_factory_id` must be per zero arg - see FS_15_79_76_85 line processor.
            self.curr_node.plugin_instance_id,
            self.curr_path,
            self.zero_arg_interp_tree,
        )
        self.visitor_decision = TraverseDecision.skip_sub_tree

    # noinspection PyPep8Naming
    def _visit_TreePathNode(self, node: TreePathNode) -> None:
        self.visitor_decision = TraverseDecision.skip_sub_tree

    # noinspection PyPep8Naming
    def _visit_InterpTreeNode(self, node: InterpTreeNode) -> None:
        self.visitor_decision = TraverseDecision.skip_sub_tree

    # noinspection PyPep8Naming
    def _visit_FuncTreeNode(self, node: FuncTreeNode) -> None:
        self.visitor_decision = TraverseDecision.skip_sub_tree


# noinspection PyPep8Naming
class _CompositeTreeWalker_jump_tree(AbstractCompositeTreeWalker):

    def __init__(
        self,
        composite_forest: CompositeForest,
        info_type: CompositeInfoType,
    ):
        # main input:
        super().__init__(
            composite_forest,
            info_type,
        )

        # results:
        self.jump_tree: dict = {}

    # noinspection PyPep8Naming
    def _visit_ZeroArgNode(self, node: ZeroArgNode) -> None:
        self.visitor_decision = TraverseDecision.walk_sub_tree

    # noinspection PyPep8Naming
    def _visit_TreePathNode(self, node: TreePathNode) -> None:
        self.visitor_decision = TraverseDecision.walk_sub_tree

    # noinspection PyPep8Naming
    def _visit_InterpTreeNode(self, node: InterpTreeNode) -> None:
        assert len(self.curr_path) > 0
        _put_value_into_dict_path(
            self.curr_node.next_interp.jump_path,
            self.curr_path,
            self.jump_tree,
        )
        self.visitor_decision = TraverseDecision.skip_sub_tree

    # noinspection PyPep8Naming
    def _visit_FuncTreeNode(self, node: FuncTreeNode) -> None:
        self.visitor_decision = TraverseDecision.skip_sub_tree


# noinspection PyPep8Naming
class _CompositeTreeWalker_interp_tree(AbstractCompositeTreeWalker):

    def __init__(
        self,
        composite_forest: CompositeForest,
        info_type: CompositeInfoType,
        plugin_instance_id: Union[str, None],
    ):
        # main input:
        super().__init__(
            composite_forest,
            info_type,
        )

        # `info_type`-specific input:
        self.plugin_instance_id: Union[str, None] = plugin_instance_id

        # internal trackers:
        self.curr_parent_plugin_instance_id: Union[str, None] = None

        # results:
        self.interp_tree: dict = {}

    # noinspection PyPep8Naming
    def _visit_ZeroArgNode(self, node: ZeroArgNode) -> None:
        if self.curr_node.plugin_instance_id == self.plugin_instance_id:
            self.curr_parent_plugin_instance_id = self.plugin_instance_id
            self.visitor_decision = TraverseDecision.walk_sub_tree
        else:
            self.visitor_decision = TraverseDecision.skip_sub_tree

    # noinspection PyPep8Naming
    def _visit_TreePathNode(self, node: TreePathNode) -> None:
        self.visitor_decision = TraverseDecision.walk_sub_tree

    # noinspection PyPep8Naming
    def _visit_InterpTreeNode(self, node: InterpTreeNode) -> None:
        if self.curr_parent_plugin_instance_id == self.plugin_instance_id:
            _put_value_into_dict_path(
                self.curr_node.plugin_instance_id,
                self.curr_path,
                self.interp_tree,
            )
        self.visitor_decision = TraverseDecision.skip_sub_tree

    # noinspection PyPep8Naming
    def _visit_FuncTreeNode(self, node: FuncTreeNode) -> None:
        self.visitor_decision = TraverseDecision.skip_sub_tree


# noinspection PyPep8Naming
class _CompositeTreeWalker_func_tree(AbstractCompositeTreeWalker):

    def __init__(
        self,
        composite_forest: CompositeForest,
        info_type: CompositeInfoType,
        plugin_instance_id: Union[str, None],
    ):
        # main input:
        super().__init__(
            composite_forest,
            info_type,
        )

        # `info_type`-specific input:
        self.plugin_instance_id: Union[str, None] = plugin_instance_id

        # internal trackers:
        self.curr_parent_plugin_instance_id: Union[str, None] = None

        # results:
        self.func_tree: dict = {}

    # noinspection PyPep8Naming
    def _visit_ZeroArgNode(self, node: ZeroArgNode) -> None:
        self.visitor_decision = TraverseDecision.walk_sub_tree

    # noinspection PyPep8Naming
    def _visit_TreePathNode(self, node: TreePathNode) -> None:
        self.visitor_decision = TraverseDecision.walk_sub_tree

    # noinspection PyPep8Naming
    def _visit_InterpTreeNode(self, node: InterpTreeNode) -> None:
        if self.curr_node.plugin_instance_id == self.plugin_instance_id:
            self.curr_parent_plugin_instance_id = self.plugin_instance_id
            self.visitor_decision = TraverseDecision.walk_sub_tree
        else:
            self.visitor_decision = TraverseDecision.skip_sub_tree

    # noinspection PyPep8Naming
    def _visit_FuncTreeNode(self, node: FuncTreeNode) -> None:
        if self.curr_parent_plugin_instance_id == self.plugin_instance_id:
            _put_value_into_dict_path(
                self.curr_node.func_id,
                self.curr_path,
                self.func_tree,
            )
        self.visitor_decision = TraverseDecision.skip_sub_tree


def extract_tree_abs_path_to_interp_id(
    composite_forest: CompositeForest,
    func_id: str,
) -> dict:
    """
    Extracts `CompositeInfoType.tree_abs_path_to_interp_id`.
    """

    tree_walker = _CompositeTreeWalker_tree_abs_path_to_interp_id(
        composite_forest,
        CompositeInfoType.tree_abs_path_to_interp_id,
        func_id,
    )
    tree_walker.walk_tree_roots()

    return tree_walker.tree_abs_path_to_interp_id


def extract_zero_arg_interp_tree(
    composite_forest: CompositeForest,
) -> dict:
    """
    Extracts `CompositeInfoType.zero_arg_interp_tree`.
    """

    tree_walker = _CompositeTreeWalker_zero_arg_interp_tree(
        composite_forest,
        CompositeInfoType.zero_arg_interp_tree,
    )
    tree_walker.walk_tree_roots()

    return tree_walker.zero_arg_interp_tree


def extract_jump_tree(
    composite_forest: CompositeForest,
) -> dict:
    """
    Extracts `CompositeInfoType.jump_tree`.
    """

    tree_walker = _CompositeTreeWalker_jump_tree(
        composite_forest,
        CompositeInfoType.jump_tree,
    )
    tree_walker.walk_tree_roots()

    return tree_walker.jump_tree


def extract_interp_tree(
    composite_forest: CompositeForest,
    plugin_instance_id: str,
) -> dict:
    """
    Extracts `CompositeInfoType.interp_tree`.
    """

    tree_walker = _CompositeTreeWalker_interp_tree(
        composite_forest,
        CompositeInfoType.interp_tree,
        plugin_instance_id,
    )
    tree_walker.walk_tree_roots()

    return tree_walker.interp_tree


def extract_func_tree(
    composite_forest: CompositeForest,
    plugin_instance_id: str,
) -> dict:
    """
    Extracts `CompositeInfoType.func_tree`.
    """

    tree_walker = _CompositeTreeWalker_func_tree(
        composite_forest,
        CompositeInfoType.func_tree,
        plugin_instance_id,
    )
    tree_walker.walk_tree_roots()

    return tree_walker.func_tree



from __future__ import annotations

import contextlib

from argrelay.composite_tree.CompositeInfoType import CompositeInfoType
from argrelay.composite_tree.CompositeTreeWalker import extract_interp_tree
from argrelay.composite_tree.DictTreeWalker import DictTreeWalker, sequence_starts_with
from argrelay.enum_desc.PluginType import PluginType
from argrelay.misc_helper_common import eprint
from argrelay.plugin_interp.AbstractInterpFactory import AbstractInterpFactory
from argrelay.plugin_interp.InterpTreeInterp import InterpTreeInterp
from argrelay.plugin_interp.InterpTreeInterpFactoryConfigSchema import (
    tree_path_interp_factory_config_desc,
    interp_selector_tree_,
)
from argrelay.runtime_context.InterpContext import InterpContext
from argrelay.runtime_data.ServerConfig import ServerConfig, assert_plugin_instance_id


class InterpTreeInterpFactory(AbstractInterpFactory):
    """
    Implements FS_01_89_09_24 interp tree.
    """

    def __init__(
        self,
        server_config: ServerConfig,
        plugin_instance_id: str,
        plugin_config_dict: dict,
    ):
        super().__init__(
            server_config,
            plugin_instance_id,
            plugin_config_dict,
        )
        self.is_recursive_func_load = False

        self._compare_config_with_composite_tree()

    # TODO_10_72_28_05: This will go away together with switch to FS_33_76_82_84 composite tree config:
    def _compare_config_with_composite_tree(
        self,
    ):
        expected_dict = self.plugin_config_dict[interp_selector_tree_]
        actual_dict = extract_interp_tree(
            self.server_config.server_plugin_control.composite_forest,
            self.plugin_instance_id,
        )
        eprint(f"expected_dict: {expected_dict}")
        eprint(f"actual_dict: {actual_dict}")
        assert expected_dict == actual_dict

    def load_config(
        self,
        plugin_config_dict,
    ) -> dict:
        return tree_path_interp_factory_config_desc.dict_from_input_dict(plugin_config_dict)

    def load_interp_tree_abs_paths(
        self,
        this_plugin_instance_interp_tree_abs_paths: list[tuple[str, ...]],
    ):
        # TODO: TODO_18_51_46_14: refactor FS_42_76_93_51 zero_arg_interp into FS_15_79_76_85 line processor:
        #       At the moment, `InterpTreeInterpFactory` is not plugged into main interp tree -
        #       it is plugged into `first_arg_vals_to_next_interp_factory_ids` of `FirstArgInterpFactory`.
        #
        #       When this func is called for `InterpTreeInterpFactory`, it cannot be found plugged inside
        #       `interp_selector_tree` - instead, it walks the `interp_selector_tree` and
        #       invokes this func for other interps in there.

        dict_tree_walker: DictTreeWalker = DictTreeWalker(
            CompositeInfoType.interp_tree,
            self.plugin_config_dict[interp_selector_tree_],
        )
        # Walk configured interp tree and call `load_interp_tree_abs_paths` with `interp_tree_abs_path` for each interp.
        all_interp_tree_abs_paths: dict[str, list[list[str]]] = dict_tree_walker.build_str_leaves_paths()
        for interp_factory_instance_id in all_interp_tree_abs_paths:
            assert_plugin_instance_id(
                self.server_config,
                interp_factory_instance_id,
                PluginType.InterpFactoryPlugin,
            )
            interp_factory: AbstractInterpFactory = self.server_config.interp_factories[interp_factory_instance_id]
            interp_factory.load_interp_tree_abs_paths([
                tuple(list_abs_path)
                for list_abs_path in all_interp_tree_abs_paths[interp_factory_instance_id]
            ])

    def load_func_envelopes(
        self,
        interp_tree_abs_path: tuple[str, ...],
        func_ids_to_func_envelopes: dict[str, dict],
    ) -> list[str]:
        with self._is_recursive_load() as is_recursive_load:
            # TODO: FS_33_76_82_84 composite tree: add validation that same interp tree interp is not loaded twice - is it required (given that plugin instances can be reused)?
            if is_recursive_load:
                return []
            return self._load_func_envelopes(
                interp_tree_abs_path,
                func_ids_to_func_envelopes,
            )

    @contextlib.contextmanager
    def _is_recursive_load(self):
        """
        Auto-reset flag `self.is_recursive_func_load` on exit from `with` section.
        """
        try:
            if not self.is_recursive_func_load:
                self.is_recursive_func_load = True
                is_recursive_load = False
            else:
                is_recursive_load = True
            yield is_recursive_load
        finally:
            self.is_recursive_func_load = False

    def _load_func_envelopes(
        self,
        interp_tree_abs_path: tuple[str, ...],
        func_ids_to_func_envelopes: dict[str, dict],
    ) -> list[str]:
        mapped_func_ids: list[str] = super().load_func_envelopes(
            interp_tree_abs_path,
            func_ids_to_func_envelopes,
        )
        dict_tree_walker: DictTreeWalker = DictTreeWalker(
            CompositeInfoType.interp_tree,
            self.plugin_config_dict[interp_selector_tree_],
        )
        # Walk configured interp tree and call `load_func_envelopes` with `interp_tree_abs_path` for each interp.
        all_interp_tree_abs_paths: dict[str, list[list[str]]] = dict_tree_walker.build_str_leaves_paths()
        for interp_factory_instance_id in all_interp_tree_abs_paths:
            for sub_interp_tree_abs_path in all_interp_tree_abs_paths[interp_factory_instance_id]:
                if not sequence_starts_with(sub_interp_tree_abs_path, interp_tree_abs_path):
                    # skip: other `interp_tree_abs_path`-s are going to be separate calls to this func:
                    continue
                assert_plugin_instance_id(
                    self.server_config,
                    interp_factory_instance_id,
                    PluginType.InterpFactoryPlugin,
                )
                interp_factory: AbstractInterpFactory = self.server_config.interp_factories[interp_factory_instance_id]

                mapped_func_ids.extend(interp_factory.load_func_envelopes(
                    tuple(sub_interp_tree_abs_path),
                    func_ids_to_func_envelopes,
                ))
        return mapped_func_ids

    def create_interp(
        self,
        interp_ctx: InterpContext,
    ) -> InterpTreeInterp:
        """
        Each `InterpTreeInterpFactory` serves a subtree,
        but it only stores config of per its plug points
        (interp tree abs paths where it was invoked on `load_func_envelopes`).
        If next interp with its plugin id is created, but the specified path is within its subtree,
        chop off the last step in the path until it matches path to its subtree to select one.
        For example:
        ```
        # Current path within interp tree:
        interp_ctx.interp_tree_abs_path = ('l1', 'l2', 'l3', 'l4', )
        # Plug point of this factory within the interp tree:
        self.interp_tree_abs_paths_to_node_configs = {
            ('l1', 'l2', ): {},
            ('x1', 'x2', 'x3', ): {},
        }
        # Keep trying to cut of the last step in the path until it matches one of the key:
        ('l1', 'l2', 'l3', 'l4', )
        ('l1', 'l2', 'l3', )
        ('l1', 'l2', )
        # Done.
        ```
        """
        orig_subtree_abs_path = interp_ctx.interp_tree_abs_path
        curr_subtree_abs_path = orig_subtree_abs_path
        while curr_subtree_abs_path not in self.interp_tree_abs_paths_to_node_configs:
            if len(curr_subtree_abs_path) == 0:
                raise ValueError(
                    f"path `{orig_subtree_abs_path}` is not subpath of `{self.interp_tree_abs_paths_to_node_configs.keys()}` paths"
                )
            curr_subtree_abs_path = curr_subtree_abs_path[:-1]
        return InterpTreeInterp(
            self.plugin_instance_id,
            self.interp_tree_abs_paths_to_node_configs[curr_subtree_abs_path],
            interp_ctx,
        )



"""
"""

goto_host_func_ = "goto_host_func"
goto_service_func_ = "goto_service_func"
goto_git_repo_func_ = "goto_git_repo_func"

desc_host_func_ = "desc_host_func"
desc_service_func_ = "desc_service_func"

list_host_func_ = "list_host_func"
list_service_func_ = "list_service_func"

diff_service_func_ = "diff_service_func"

desc_git_tag_func_ = "desc_git_tag_func"
desc_git_commit_func_ = "desc_git_commit_func"



import os
from typing import Union

from argrelay.misc_helper_common import get_argrelay_dir
from argrelay.misc_helper_common.TypeDesc import TypeDesc


def load_client_plugin_config(
    plugin_instance_id: str,
    type_desc: TypeDesc,
) -> Union[dict, None]:
    """
    This function loads plugin config on client side.

    See FS_83_23_99_90 client plugin config override.
    """

    config_file_path = f"{get_argrelay_dir()}/conf/plugin_config/{plugin_instance_id}.yaml"
    if os.path.isfile(config_file_path):
        return type_desc.dict_from_yaml_file(config_file_path)
    else:
        return None



from __future__ import annotations

from copy import deepcopy
from dataclasses import field, dataclass
from typing import Union

from argrelay.enum_desc.ArgSource import ArgSource
from argrelay.enum_desc.InterpStep import InterpStep
from argrelay.enum_desc.ServerAction import ServerAction
from argrelay.enum_desc.SpecialChar import SpecialChar
from argrelay.enum_desc.TermColor import TermColor
from argrelay.misc_helper_common import eprint
from argrelay.misc_helper_common.ElapsedTime import ElapsedTime
from argrelay.relay_server.HelpHintCache import HelpHintCache
from argrelay.relay_server.QueryEngine import QueryEngine, populate_query_dict
from argrelay.relay_server.QueryResult import QueryResult
from argrelay.runtime_context.EnvelopeContainer import EnvelopeContainer
from argrelay.runtime_context.ParsedContext import ParsedContext
from argrelay.runtime_context.SearchControl import SearchControl
from argrelay.runtime_context.arg_buckets_utils import arg_buckets_to_token_ipos_list

function_container_ipos_ = 0


@dataclass
class InterpContext:
    """
    Mutable state for the process of command line interpretation.
    """

    parsed_ctx: ParsedContext = field()

    # TODO: Move all dynamic and non-serializable objects into `InterpRuntime` (or something like that).
    interp_factories: dict[str, "AbstractInterpFactory"] = field()
    """
    Reference to `ServerConfig.action_delegators`.
    """

    action_delegators: dict[str, "AbstractDelegator"] = field()
    """
    Reference to `ServerConfig.action_delegators`.
    """

    query_engine: QueryEngine = field()

    help_hint_cache: HelpHintCache = field()

    excluded_tokens: list[int] = field(init = False, default_factory = lambda: [])
    """
    Tokens excluded by ways other than consumption into `consumed_arg_buckets`.
    """

    included_arg_buckets: list[list[int]] = field(init = False, default_factory = lambda: [])
    """
    FS_97_64_39_94: arg buckets
    
    If `included_arg_buckets` are combined,
    the result will contain both `remaining_arg_buckets` and `consumed_arg_buckets`.
    Field `included_arg_buckets` is maximum set for arg buckets -
    it is similar to maximum set `ParsedContext.all_tokens`,
    but it cannot contain all tokens - it must exclude at least `SpecialChar.ArgBucketDelimiter` to start with.
    """

    remaining_arg_buckets: list[list[int]] = field(init = False)
    """
    Same as `included_arg_buckets` but for remaining tokens only.
    """

    consumed_arg_buckets: list[list[int]] = field(init = False, default_factory = lambda: [])
    """
    Same as `included_arg_buckets` but for consumed tokens only.
    """

    token_ipos_to_arg_bucket_map: dict[int, int] = field(init = False, default_factory = lambda: {})
    """
    Index reversed to `included_arg_buckets` -
    for each `token_ipos` it gives the index of the `arg_bucket` it is in.
    """

    envelope_containers: list[EnvelopeContainer] = field(init = False, default_factory = lambda: [])
    """
    Each `envelope_container` wraps `data_envelope`-s matching query and some associated data.
    """

    curr_container: EnvelopeContainer = field(init = False, default = None)
    """
    One of the `envelope_containers` currently being searched.
    """

    curr_container_ipos: int = field(init = False, default = -1)
    """
    It is an `ipos` into `envelope_containers` to select `curr_container` (what is currently being searched for).

    Check `EnvelopeContainer.found_count` to see whether this container is singled out or not.

    All `envelope_containers` after `curr_container_ipos` will not be pointing to any `data_envelope`-s
    (search procedure has not reached them yet).
    """

    prev_interp: "AbstractInterp" = field(init = False, default = None)

    curr_interp: "AbstractInterp" = field(init = False, default = None)
    """
    Current interpreter during command line interpretation.
    """

    comp_suggestions: list[str] = field(init = False, default_factory = lambda: [])
    """
    Sorted list of completion suggestions.
    """

    interp_tree_abs_path: tuple[str, ...] = field(init = False, default = tuple([]))
    """
    Provides curr path within FS_01_89_09_24 interp tree.
    Takes part in implementation of FS_01_89_09_24 interp tree.
    """

    def __post_init__(self):
        self._init_arg_buckets()

    def _init_arg_buckets(self):
        self.included_arg_buckets.append([])
        curr_bucket_index = len(self.included_arg_buckets) - 1
        curr_arg_bucket = self.included_arg_buckets[curr_bucket_index]
        for token_ipos in range(0, len(self.parsed_ctx.all_tokens)):

            # Split and populate FS_97_64_39_94 arg buckets:
            if self.parsed_ctx.all_tokens[token_ipos] == SpecialChar.ArgBucketDelimiter.value:
                self.included_arg_buckets.append([])
                curr_bucket_index = len(self.included_arg_buckets) - 1
                curr_arg_bucket = self.included_arg_buckets[curr_bucket_index]
                # Exclude `SpecialChar.ArgBucketDelimiter` unconditionally:
                self.excluded_tokens.append(token_ipos)
            else:

                if self.parsed_ctx.server_action is ServerAction.ProposeArgValues:
                    # FS_23_62_89_43 tangent arg value completion:
                    # `ServerAction.ProposeArgValues` excludes tangent token because it is supposed to be completed:
                    if token_ipos == self.parsed_ctx.tan_token_ipos:
                        self.excluded_tokens.append(token_ipos)
                    else:
                        curr_arg_bucket.append(token_ipos)
                        self.token_ipos_to_arg_bucket_map[token_ipos] = curr_bucket_index
                else:
                    # FS_23_62_89_43 tangent arg value completion:
                    # Process all tokens (including tangent token) in case of `ServerAction.DescribeLineArgs`.
                    # Obviously, same applies for `ServerAction.RelayLineArgs` (as there is no chance to propose more).
                    curr_arg_bucket.append(token_ipos)
                    self.token_ipos_to_arg_bucket_map[token_ipos] = curr_bucket_index

        # Init remaining and consumed:
        self.remaining_arg_buckets = deepcopy(self.included_arg_buckets)
        for i in range(len(self.included_arg_buckets)):
            self.consumed_arg_buckets.append([])

    def next_remaining_token_ipos(
        self,
    ) -> Union[int, None]:
        for arg_bucket in self.remaining_arg_buckets:
            for token_ipos in arg_bucket:
                return token_ipos
        return None

    def remaining_token_ipos_list(
        self,
    ) -> list[int]:
        return arg_buckets_to_token_ipos_list(self.remaining_arg_buckets)

    def consumed_token_ipos_list(
        self,
    ) -> list[int]:
        return arg_buckets_to_token_ipos_list(self.consumed_arg_buckets)

    def alloc_searchable_containers(
        self,
        search_control_list: list[SearchControl],
    ):
        for search_control in search_control_list:
            envelope_container = EnvelopeContainer(search_control)
            self.envelope_containers.append(envelope_container)

    def is_last_container(self) -> bool:
        return self.curr_container_ipos + 1 == len(self.envelope_containers)

    def is_func_found(self) -> bool:
        return (
            self.curr_container_ipos >= function_container_ipos_
            and
            self.envelope_containers[function_container_ipos_].found_count == 1
        )

    def query_prop_values(self):
        if not (
            self.curr_container
            and
            self.curr_container.search_control.envelope_class
        ):
            # If `ReservedPropName.envelope_class` is not specified, nothing to search:
            return

        ElapsedTime.measure(f"begin_query_envelopes: {self.curr_container.search_control.envelope_class}")
        query_dict = populate_query_dict(self.curr_container)
        query_result: QueryResult = self.query_engine.query_prop_values(
            query_dict,
            self.curr_container.search_control,
            self.curr_container.assigned_types_to_values,
        )
        self.curr_container.remaining_types_to_values = query_result.remaining_types_to_values
        self.curr_container.data_envelopes = query_result.data_envelopes
        self.curr_container.found_count = query_result.found_count
        ElapsedTime.measure(f"end_query_envelopes: {query_dict} {self.curr_container.found_count}")

    def consume_args(
        self,
        interp_n: int
    ):
        while True:

            # FS_44_36_84_88 consume args one by one:
            while True:
                # Query envelope values only - they will be used for consumption of command line args:
                ElapsedTime.measure(f"[i={interp_n}]: before_entry_query: {self.curr_interp}")
                self.query_prop_values()
                # Reset to False as we just executed new query:
                arg_was_consumed = False

                # Because each `prop_value` set (per `prop_type`) is treated independently,
                # assignment of `prop_value`-s from args may set combinations
                # which yields no search result in subsequent query.
                # Logically, it is better to consume args one by one and running query after each consumption,
                # but this is not done due to query overhead.
                # Note that Tab-completion and selection (via manual step by human) in separate requests to server and
                # separate `interpret_command` calls is close to that logically better approach.
                if not arg_was_consumed:
                    ElapsedTime.measure(f"[i={interp_n}]: before_consume_key_args: {self.curr_interp}")
                    arg_was_consumed = self.curr_interp.consume_key_args()
                if not arg_was_consumed:
                    ElapsedTime.measure(f"[i={interp_n}]: before_consume_pos_args: {self.curr_interp}")
                    arg_was_consumed = self.curr_interp.consume_pos_args()

                if arg_was_consumed:
                    if self.curr_interp.consumes_args_at_once():
                        # No known interp consuming at once depends on query - do not query, simply exit loop:
                        break
                    else:
                        # Run next cycle to see if one more can be consumed:
                        pass
                else:
                    break

            query_changed = False

            if self.curr_container:
                # Set implicit values (so that applying defaults knows what they are):
                self.curr_container.populate_implicit_arg_values()

            if self.curr_interp.has_fill_control():
                # FS_72_53_55_13: options before defaults
                # Describing args will need to show options except default - query values before applying defaults:
                ElapsedTime.measure(f"[i={interp_n}]: before_query_without_defaults: {self.curr_interp}")
                # This step may be optimized away and is only needed to detect non-default options for
                # `ServerAction.DescribeLineArgs` but we perform it for all requests to
                # ensure they have common view in case of possible data issues
                # (e.g. TODO_39_25_11_76: `data_envelope`-s with missing props).
                self.query_prop_values()
                # Reset to False as we just executed new query:
                query_changed = False

                self._save_potentially_hidden_by_defaults()

                # Apply defaults (they may apply more than single value at a time):
                query_changed = (
                    self.curr_interp.delegate_fill_control()
                    or
                    query_changed
                )
                self._leave_only_hidden_by_defaults()

            # Query envelopes after all implicit and default values assigned:
            # TODO: We could probably select whether to query only envelopes or
            #       query their values depending on `ServerAction`.
            #       But init of next envelope depends on prev envelope found.
            if query_changed:
                if self.curr_interp.consumes_args_at_once():
                    # No known interp consuming at once depends on query - do not query, simply exit loop:
                    break
                else:
                    pass
            else:
                break

    # TODO: Move all dynamic and non-serializable objects into `InterpRuntime` (or something like that) together with this loop:
    def interpret_command(
        self,
        first_interp_factory_id: str,
    ) -> None:
        """
        Implements FS_55_57_45_04 enum selector version of FS_15_79_76_85 line processor.

        Start with initial interpreter and continue until curr interpreter returns no more next interpreter.

        TODO: TODO_18_51_46_14: refactor FS_42_76_93_51 zero_arg_interp into FS_15_79_76_85 line processor.
        """

        interp_n: int = 0

        # FS_01_89_09_24 interp tree:
        # start of tracking interp tree abs path while processing command
        # (to pass it to the next interp factory to create next interp):
        self.interp_tree_abs_path = tuple([])

        self.curr_interp = self.create_next_interp(first_interp_factory_id)
        while True:
            interp_n += 1

            self.consume_args(interp_n)

            ElapsedTime.measure(f"[i={interp_n}]: before_try_iterate: {self.curr_interp}")
            interp_step: InterpStep = self.curr_interp.try_iterate()
            ElapsedTime.measure(f"[i={interp_n}]: after_try_iterate: {self.curr_interp}: {interp_step}")
            if interp_step is InterpStep.NextEnvelope:
                continue
            elif interp_step is InterpStep.StopAll:
                ElapsedTime.measure(
                    f"[i={interp_n}]: before_contribute_to_completion: {self.curr_interp}"
                )
                self._contribute_to_completion()
                ElapsedTime.measure(
                    f"[i={interp_n}]: after_contribute_to_completion: {self.curr_interp}"
                )
                return
            elif interp_step is InterpStep.NextInterp:
                ElapsedTime.measure(
                    f"[i={interp_n}]: before_contribute_to_completion: {self.curr_interp}"
                )
                self._contribute_to_completion()
                ElapsedTime.measure(
                    f"[i={interp_n}]: after_contribute_to_completion: {self.curr_interp}"
                )
                self.prev_interp = self.curr_interp
                next_interp = self.curr_interp.next_interp()
                if next_interp:
                    self.curr_interp = next_interp
                else:
                    return
            else:
                raise RuntimeError(interp_step)

    def _save_potentially_hidden_by_defaults(self):
        """
        Save all `remaining_types_to_values` before applying defaults.

        These `filled_types_to_values_hidden_by_defaults` are subsequently
        filtered out in `_leave_only_hidden_by_defaults`.
        """
        if not self.curr_container:
            return
        for remaining_type in self.curr_container.remaining_types_to_values.keys():
            self.curr_container.filled_types_to_values_hidden_by_defaults[
                remaining_type
            ] = deepcopy(self.curr_container.remaining_types_to_values[remaining_type])

    def _leave_only_hidden_by_defaults(self):
        """
        Weed out those `assigned_types_to_values` from `filled_types_to_values_hidden_by_defaults`
        which were set by applying defaults.

        Use `filled_types_to_values_hidden_by_defaults` saved before applying defaults
        (in `_save_potentially_hidden_by_defaults` for all `remaining_types_to_values`)
        and remove those which were not hidden by defaults
        to yield only those hidden by defaults.
        """
        if not self.curr_container:
            return
        types_not_hidden_by_defaults = []
        for type_potentially_hidden_by_defaults in self.curr_container.filled_types_to_values_hidden_by_defaults.keys():
            if type_potentially_hidden_by_defaults in self.curr_container.assigned_types_to_values:
                if (
                    self.curr_container.assigned_types_to_values[
                        type_potentially_hidden_by_defaults
                    ].arg_source
                    is not
                    ArgSource.DefaultValue
                ):
                    types_not_hidden_by_defaults.append(type_potentially_hidden_by_defaults)
            else:
                types_not_hidden_by_defaults.append(type_potentially_hidden_by_defaults)
        # Delete items outside the previous loop:
        for arg_type in types_not_hidden_by_defaults:
            del self.curr_container.filled_types_to_values_hidden_by_defaults[arg_type]

    def _contribute_to_completion(self):
        if self.parsed_ctx.server_action is ServerAction.ProposeArgValues:
            # Each in the chains of interpreters hava a chance to suggest completion values (contribute):
            self.curr_interp.propose_arg_completion()

    def propose_arg_values(self) -> list[str]:
        """
        FS_71_87_33_52: remove `help_hint` (after first space) if there is only one option
        """
        if len(self.comp_suggestions) == 1:
            first_space_ipos = self.comp_suggestions[0].find(' ')
            if first_space_ipos >= 0:
                self.comp_suggestions[0] = self.comp_suggestions[0][:first_space_ipos]
        return self.comp_suggestions

    def create_next_interp(
        self,
        next_interp_factory_id: str,
    ) -> "AbstractInterp":
        interp_factory: "AbstractInterpFactory" = self.interp_factories[next_interp_factory_id]
        return interp_factory.create_interp(
            self,
        )

    def print_debug(self, end_str: str = "\n") -> None:
        if not self.parsed_ctx.is_debug_enabled:
            return
        self.parsed_ctx.print_debug("")
        eprint(TermColor.debug_output.value, end = "")
        eprint(f"comp_suggestions: {self.comp_suggestions}", end = " ")
        eprint(TermColor.reset_style.value, end = end_str)



from __future__ import annotations

import string
from collections.abc import Callable
from copy import deepcopy
from typing import Union, Sequence

from argrelay.composite_tree.CompositeInfoType import CompositeInfoType

surrogate_node_id_ = ""
"""
Surrogate node id allows "planting" entire sub-tree in place of itself.

But the node that this surrogate node id identifies is not removed (the tree is not modified) -
functionally that node must be present. It is only excluded from the path to access its sub-tree.

Surrogate node id is skipped and does not affect (e.g.) command line
(as trees often used to construct path which would be used to select its nodes via command line).

Surrogate node id can also simplify representation of trees with leaf value only
(as replacement of `dict` type where leaf type is used directly), for example,
instead of:
```yaml
some_tree:
    "": leaf_value
```
use this (no `dict`):
```yaml
some_tree: leaf_value
```
"""

surrogate_tree_leaf_ = ""
"""
Surrogate tree leaf can be used as `dict` value to instruct to use `dict` key instead of `dict` value.

For example, this trees are equivalent:
```yaml
some_tree:
    "some_value": ""
```
use this (no `dict`):
```yaml
some_tree: "some_value"
```
"""


def normalize_tree(
    input_tree: Union[dict, str],
) -> dict:
    """
    Transform `input_tree` to some equivalent form which makes all trees comparable.

    *   If `input_tree` is not `dict`, make it `dict` by inserting using `surrogate_node_id_`.
    *   If `input_tree` has only one node with `surrogate_node_id_`, collapse that layer.
    *   Tree of 0 depth with (non-`dict`) is transformed to `dict` with `surrogate_tree_leaf_`.

    NOTE: `input_tree` of depth 0 with `str` leave equal to `surrogate_tree_leaf_` is invalid.

    NOTE: These two dicts are equivalent:
    *   with `surrogate_tree_leaf_`: { "some_value": "" }
    *   with `surrogate_node_id_`: { "": "some_value" }
    To resolve the ambiguity, "some_value" is transformed using `surrogate_node_id_`.
    """

    return _normalize_tree(input_tree, 0)


def _normalize_tree(
    input_tree: Union[dict, str],
    tree_depth: int,
) -> Union[dict, str]:
    output_tree: Union[dict, str] = {}

    if isinstance(input_tree, str):
        if input_tree == surrogate_tree_leaf_:
            if tree_depth != 0:
                return input_tree
            else:
                raise ValueError(f"`input_tree` with single leaf `{surrogate_node_id_}` is invalid.")
        else:
            if tree_depth != 0:
                return input_tree
            else:
                output_tree[surrogate_node_id_] = input_tree
    elif isinstance(input_tree, dict):
        if len(input_tree) == 1 and surrogate_node_id_ in input_tree:
            # collapse `dict` to `str`:
            output_tree = input_tree[surrogate_node_id_]
        else:
            for node_id in input_tree:
                output_tree[node_id] = _normalize_tree(input_tree[node_id], tree_depth + 1)
    else:
        raise ValueError(f"unexpected `input_tree` type: {type(input_tree)}")

    return output_tree


def fetch_subtree_node(
    tree_dict: dict,
    node_path: Sequence[str],
):
    """
    Fetches tree node by node path.
    """

    curr_node_value = tree_dict
    for node_path_part in node_path:
        if isinstance(curr_node_value, dict) and node_path_part in curr_node_value:
            curr_node_value = curr_node_value[node_path_part]
        else:
            return None
    return curr_node_value


def sequence_starts_with(
    outer_seq: Sequence[str],
    inner_seq: Sequence[str],
) -> bool:
    """
    Similar to `str.startwith` but in terms of sequence elements.
    """
    for part_index, part_value in enumerate(inner_seq):
        if part_index < len(outer_seq):
            if outer_seq[part_index] != part_value:
                return False
        else:
            # `outer_seq` is smaller than `inner_seq`:
            return False
    return True


def contains_whitespace(s):
    for c in s:
        if c in string.whitespace:
            return True
    return False


class DictTreeWalker:
    """
    This tree walker is used to build various data structures extracted from (simple) `dict` trees:
    *   FS_01_89_09_24 interp tree
    *   FS_26_43_73_72 func tree
    *   FS_91_88_07_23 jump tree

    As opposed to (complex) FS_33_76_82_84 `composite_tree` (with node types),
    the `dict` trees are specified in config via `dict` without node types -
    only node id names (as `dict` keys) and tree leaves (as `str` or `list`).

    These simple trees can also be extracted from FS_33_76_82_84 `composite_tree` via `CompositeWalker`
    (where each node has its type and additional info, not just path step name and value at tree leaves).
    """

    def __init__(
        self,
        info_type: CompositeInfoType,
        tree_dict: dict,
    ):
        self.info_type = info_type
        self.tree_dict = tree_dict
        self.leaf_type = None

    def build_str_leaves_paths(
        self,
    ) -> dict[str, list[list[str]]]:
        """
        For each named leaf (becomes key) provide its paths in the tree.

        Each named leaf can have several paths in the tree (leaves with the same name).

        Use case:
        *   We know `func_id`, we need to know which FS_26_43_73_72 func tree paths it is registered under.
        """
        tree_paths: dict[str, list[list[str]]] = {}
        self.leaf_type = str
        self._build_leaves_sub_paths(
            [],
            self.tree_dict,
            tree_paths,
            self._handle_str_leaf,
        )
        return tree_paths

    # TODO: Unused (and, apparently, it wasn't used in the past).
    def build_tuple_leaves_paths(
        self,
    ) -> dict[tuple[str, ...], list[list[str]]]:
        """
        For each leaf value (`list` in tree leaf becomes `tuple` key in the map) provide its paths in the tree.

        Each leaf (its `tuple` converted from `list`) can have several paths in the tree (leaves with the same value).

        In tree, leaf is a `list`.
        In map, leaf becomes `tuple` (as a key for `dict`).

        Use case:
        *   We know tree path (tuple) defined in the nodes, we need to know all tree paths where same path is defined.
            (not a real use case as it is unused TODO: clean up)
        """
        tree_paths: dict[tuple[str, ...], list[list[str]]] = {}
        self.leaf_type = list
        self._build_leaves_sub_paths(
            [],
            self.tree_dict,
            tree_paths,
            self._handle_list_leaf,
        )
        return tree_paths

    def build_paths_to_paths(
        self
    ) -> dict[tuple[str, ...], tuple[str, ...]]:
        """
        Map each path in the tree into path specified in the leaf (`list` in tree leaf becomes `tuple` key in the map).

        In tree, leaf is a `list`.
        In map, leaf becomes `tuple` (as a key for `dict`).

        Use case:
        *   We know curr tree path (tuple), we need to know next tree path where to jump (FS_91_88_07_23 jump tree).
        """

        paths_to_paths: dict[tuple[str, ...], tuple[str, ...]] = {}
        self.leaf_type = list
        self._build_paths_to_paths(
            [],
            self.tree_dict,
            paths_to_paths,
        )
        return paths_to_paths

    def _build_leaves_sub_paths(
        self,
        curr_path: list[str],
        sub_tree: Union[str, list, dict],
        output_leaves_paths: dict[Union[str, tuple[str, ...]], list[list[str]]],
        handle_leaf: Callable[
            [
                list[str],
                Union[str, list, dict],
                dict[Union[str, tuple[str, ...]],
                list[list[str]]],
            ],
            None,
        ],
    ):
        if isinstance(sub_tree, dict):
            for node_id in sub_tree:
                tree_node = sub_tree[node_id]
                if node_id == surrogate_node_id_:
                    self._build_leaves_sub_paths(
                        curr_path,
                        tree_node,
                        output_leaves_paths,
                        handle_leaf,
                    )
                else:
                    next_leaf_path = deepcopy(curr_path)
                    next_leaf_path.append(node_id)
                    self._build_leaves_sub_paths(
                        next_leaf_path,
                        tree_node,
                        output_leaves_paths,
                        handle_leaf,
                    )
        elif isinstance(sub_tree, self.leaf_type):
            handle_leaf(
                curr_path,
                sub_tree,
                output_leaves_paths,
            )
        else:
            self._raise_on_wrong_type(curr_path, sub_tree)

    def _build_paths_to_paths(
        self,
        curr_path: list[str],
        sub_tree: Union[list, dict],
        output_paths_to_paths: dict[tuple[str, ...], tuple[str, ...]],
    ):
        if isinstance(sub_tree, dict):
            for node_id in sub_tree:
                tree_node = sub_tree[node_id]
                if node_id == surrogate_node_id_:
                    self._build_paths_to_paths(
                        curr_path,
                        tree_node,
                        output_paths_to_paths,
                    )
                else:
                    next_leaf_path = deepcopy(curr_path)
                    next_leaf_path.append(node_id)
                    self._build_paths_to_paths(
                        next_leaf_path,
                        tree_node,
                        output_paths_to_paths,
                    )
        elif isinstance(sub_tree, self.leaf_type):
            self._verify_list_str(curr_path, sub_tree)
            output_paths_to_paths[tuple(curr_path)] = tuple(sub_tree)
        else:
            self._raise_on_wrong_type(curr_path, sub_tree)

    def _handle_str_leaf(
        self,
        curr_path: list[str],
        leaf_node: str,
        output_leaves_paths: dict[str, list[list[str]]],
    ) -> None:
        if leaf_node == surrogate_tree_leaf_:
            if len(curr_path) < 1:
                raise ValueError(
                    f"{self.info_type}: tree path `{curr_path}` must contain at least one step to use leaf `{surrogate_tree_leaf_}`"
                )
            leaf_id = curr_path[-1]
            leaf_path = curr_path[:-1]
        else:
            leaf_id = leaf_node
            leaf_path = deepcopy(curr_path)

        if leaf_id in output_leaves_paths:
            output_leaves_paths[leaf_id].append(leaf_path)
        else:
            output_leaves_paths[leaf_id] = [leaf_path]

    def _handle_list_leaf(
        self,
        curr_path: list[str],
        leaf_node: list,
        output_leaves_paths: dict[tuple[str, ...], list[list[str]]],
    ) -> None:
        """
        Convert `list` leaf node to `tuple` to be used as key in `output_leaves_paths` `dict`.
        """
        self._verify_list_str(curr_path, leaf_node)

        leaf_id = tuple(leaf_node)
        leaf_path = deepcopy(curr_path)

        if leaf_id in output_leaves_paths:
            output_leaves_paths[leaf_id].append(leaf_path)
        else:
            output_leaves_paths[leaf_id] = [leaf_path]

    def _verify_list_str(
        self,
        curr_path: list[str],
        list_value: list,
    ):
        if not all(isinstance(leaf_list_item, str) for leaf_list_item in list_value):
            raise ValueError(
                f"{self.info_type}: tree path `{curr_path}` leaf list must contain only `str`"
            )

    def _raise_on_wrong_type(
        self,
        curr_path: list[str],
        typed_value,
    ):
        raise ValueError(
            f"{self.info_type}: tree path `{curr_path}` leaf type `{typed_value.__class__.__name__}` is neither `{self.leaf_type.__name__}` nor `dict`"
        )



from enum import Enum, auto


class ReservedPropName(Enum):
    """
    Some of the `data_envelope` props used to implement various `argrelay` features.
    """

    envelope_class = auto()

    func_id = auto()
    """
    Used by FS_26_43_73_72 func tree to distinguish leaves (func ids).
    It is normally the same with `FunctionEnvelopeInstanceDataSchema.func_id_`.
    """

    # TODO: Review: how it is used. Should it be renamed to `prop_name`? Does it make sense to have a property named as `prop_name`?
    arg_type = auto()

    # TODO: Review: how it is used. Should it be renamed to `prop_value`? Does it make sense to have a property named as `prop_value`?
    arg_value = auto()

    help_hint = auto()

    func_state = auto()



"""
Python part of FS_36_17_84_44 `check_env` implementation.
"""

import os
import sys

from argrelay import misc_helper_common
from argrelay.check_env.CheckEnvResult import CheckEnvResult
from argrelay.check_env.PluginCheckEnvAbstract import PluginCheckEnvAbstract
from argrelay.enum_desc.PluginSide import PluginSide
from argrelay.enum_desc.PluginType import PluginType
from argrelay.enum_desc.ResultCategory import ResultCategory
from argrelay.enum_desc.TermColor import TermColor
from argrelay.misc_helper_common import eprint
from argrelay.runtime_context.PluginClientAbstract import instantiate_client_plugin
from argrelay.runtime_data.PluginConfig import PluginConfig
from argrelay.runtime_data.PluginEntry import PluginEntry
from argrelay.schema_config_plugin.PluginConfigSchema import plugin_config_desc

# Color scheme has to be synced with `@/exe/check_env.bash`:
success_color = TermColor.back_dark_green.value
warning_color = TermColor.back_dark_yellow.value
failure_color = TermColor.back_dark_red.value
field_color = TermColor.fore_bright_cyan.value
failure_message = TermColor.fore_dark_red.value
warning_message = TermColor.fore_bright_yellow.value
success_message = TermColor.fore_bright_green.value
reset_style = TermColor.reset_style.value


# TODO: TODO_67_33_03_53.add_check_env_test_support.md
def main():
    argrelay_dir: str = os.path.realpath(os.path.abspath(sys.argv[1]))
    misc_helper_common.set_argrelay_dir(argrelay_dir)

    plugin_config: PluginConfig = plugin_config_desc.obj_from_default_file()

    total_success: bool = True

    for plugin_instance_id in plugin_config.plugin_instance_id_activate_list:
        plugin_entry: PluginEntry = plugin_config.plugin_instance_entries[plugin_instance_id]

        if not plugin_entry.plugin_enabled:
            continue

        if (
            plugin_entry.plugin_side != PluginSide.PluginClientSideOnly
            and
            plugin_entry.plugin_side != PluginSide.PluginAnySide
        ):
            continue

        plugin_instance: PluginCheckEnvAbstract = instantiate_client_plugin(
            plugin_instance_id,
            plugin_entry,
        )
        plugin_type = plugin_instance.get_plugin_type()

        if plugin_type is PluginType.CheckEnvPlugin:
            plugin_instance: PluginCheckEnvAbstract
            plugin_instance.activate_plugin()

        check_env_results: list[CheckEnvResult] = plugin_instance.execute_check()
        for check_env_result in check_env_results:
            is_failure: bool = (
                check_env_result.result_category is ResultCategory.ExecutionFailure
                or
                check_env_result.result_category is ResultCategory.VerificationFailure
            )
            is_warning: bool = check_env_result.result_category is ResultCategory.VerificationWarning

            total_success = total_success and not is_failure

            # Print level:
            if is_failure:
                eprint(f"{failure_color}ERROR:{reset_style}", end = " ")
            else:
                if is_warning:
                    eprint(f"{warning_color}WARN:{reset_style}", end = " ")
                else:
                    eprint(f"{success_color}INFO:{reset_style}", end = " ")

            # Print field:
            if check_env_result.result_key is not None:
                eprint(f"{field_color}{check_env_result.result_key}:{reset_style}", end = " ")
            else:
                pass

            # Print value:
            if check_env_result.result_value is not None:
                eprint(f"{check_env_result.result_value}", end = " ")
            else:
                pass

            # Print message:
            if check_env_result.result_message is not None:
                if is_failure:
                    eprint(f"{failure_message}# {check_env_result.result_message}{reset_style}", end = " ")
                else:
                    if is_warning:
                        eprint(f"{warning_message}# {check_env_result.result_message}{reset_style}", end = " ")
                    else:
                        eprint(f"{success_message}# {check_env_result.result_message}{reset_style}", end = " ")
            else:
                pass

            # Terminate line:
            eprint()

    if total_success:
        exit(0)
    else:
        exit(1)


if __name__ == "__main__":
    main()



from marshmallow import Schema, RAISE, fields

from argrelay.misc_helper_common.TypeDesc import TypeDesc

single_func_id_ = "single_func_id"
# TODO_10_72_28_05: This should be removed with FS_33_76_82_84 composite tree:
tree_abs_path_to_interp_id_ = "tree_abs_path_to_interp_id"


# TODO_40_10_18_32: add custom base to all schemas:
class AbstractJumpDelegatorConfigSchema(Schema):
    class Meta:
        unknown = RAISE
        strict = True

    single_func_id = fields.String(
        required = True,
    )

    tree_abs_path_to_interp_id = fields.Dict(
        keys = fields.String(),
        # TODO_79_67_28_83: Express recursive dict schema:
        # This is a tree (`dict`) of arbitrary depth with `str` leaves.
        # Ideally, this should be defined as nested `dict`,
        # but it is unknown how to do it in marshmallow.
        values = fields.Raw(),
        required = True,
    )


abstract_jump_delegator_config_example = {
    single_func_id_: "func_id_intercept_invocation",
    tree_abs_path_to_interp_id_: {
        "relay_demo": {
            "intercept": "InterpTreeInterpFactory.default",
            "duplicates": {
                "intercept": "InterpTreeInterpFactory.default",
            },
        },
        "some_command": {
            "intercept": "InterpTreeInterpFactory.default",
            "duplicates": {
                "intercept": "InterpTreeInterpFactory.default",
            },
        },
    },
}

abstract_jump_delegator_config_desc = TypeDesc(
    dict_schema = AbstractJumpDelegatorConfigSchema(),
    ref_name = AbstractJumpDelegatorConfigSchema.__name__,
    dict_example = abstract_jump_delegator_config_example,
    default_file_path = "",
)



from enum import Enum, auto


class CheckEnvFunc(Enum):
    """
    Functions used by FS_36_17_84_44 `check_env`.
    """

    func_id_get_server_project_git_commit_id = auto()
    """
    TODO
    """

    func_id_get_server_start_time = auto()
    """
    TODO
    """

    # TODO: add func to get server version



from enum import Enum

from argrelay.enum_desc.PluginSide import PluginSide


class PluginType(Enum):
    """
    See `AbstractPlugin.get_plugin_type`.
    """

    LoaderPlugin = PluginSide.PluginServerSideOnly
    """
    See classes derived from `AbstractLoader`.
    """

    InterpFactoryPlugin = PluginSide.PluginServerSideOnly
    """
    See classes derived from `AbstractInterpFactory`.
    """

    DelegatorPlugin = PluginSide.PluginServerSideOnly
    """
    See classes derived from `AbstractDelegator`.

    Note that this plugin also implements client-side logic
    but it is not instantiated on the client-side -
    instead, static method is called directly on the plugin class.
    """

    ConfiguratorPlugin = PluginSide.PluginServerSideOnly
    """
    See `AbstractConfigurator`.
    """

    CheckEnvPlugin = PluginSide.PluginClientSideOnly
    """
    See classes derived from `PluginCheckEnvAbstract`.
    """

    def __new__(
        cls,
        plugin_side: PluginSide,
    ):
        enum_value = len(cls.__members__) + 1
        enum_obj = object.__new__(cls)
        enum_obj._value_ = enum_value
        return enum_obj

    def __init__(
        self,
        plugin_side: PluginSide,
    ):
        self.plugin_side = plugin_side

    def __str__(self):
        return self.name



from marshmallow import Schema, fields, validates_schema, INCLUDE, post_dump

from argrelay.enum_desc.ReservedEnvelopeClass import ReservedEnvelopeClass
from argrelay.enum_desc.ReservedPropName import ReservedPropName
from argrelay.misc_helper_common.TypeDesc import TypeDesc
from argrelay.schema_config_interp.FunctionEnvelopeInstanceDataSchema import function_envelope_instance_data_desc

mongo_id_ = "_id"
"""
Field used by Mongo DB to uniquely identify every object inside a collection.
If not provided, it is auto-generated as `bson.objectid.ObjectId`.
Note that it is not serializable by `Schema.dump` and must be excluded.
for `argrelay`, if unique id is required, `envelope_id_` must be used (which is copied to `_id` on load).
"""

envelope_id_ = "envelope_id"
"""
Not required field with unique id for `data_envelope`.
If provided, it is given to MongoDB as `mongo_id_` (otherwise, if not provided, MongoDB auto-generates one).
"""

# TODO_45_75_75_65: Merge `instance_data` into `envelop_payload`:
instance_data_ = "instance_data"
"""
Data specific to `envelope_class`.
Unlike `envelope_payload` `argrelay` does not inspect, `instance_data` can be inspected
(if not inspected by `argrelay`, but by its plugins) and this data has schema implied or defined somewhere.
"""

envelope_payload_ = "envelope_payload"
"""
Data `argrelay` does not inspect.
"""


class DataEnvelopeSchema(Schema):
    """
    Schema for all :class:`StaticDataSchema.data_envelopes`
    """

    class Meta:
        # All other fields of data envelope becomes its metadata props available for search queries.
        # Note that it does not work for `Schema.dump`, only on `Schema.load`
        # (see `keep_unknown_fields` func below to work around that):
        unknown = INCLUDE
        strict = True

    envelope_id = fields.String(
        required = False,
    )

    """
    Data specific to `envelope_class`.
    Each envelope class may define its own schema for that data.
    For example, `ReservedEnvelopeClass.ClassFunction` defines `FunctionEnvelopeInstanceDataSchema`.
    """
    # TODO_45_75_75_65: Merge `instance_data` into `envelop_payload`:
    instance_data = fields.Dict(
        required = False,
    )

    """
    Arbitrary schemaless data (payload) wrapped by `DataEnvelopeSchema`.
    It is not inspected by `argrelay` layer (unlike, `instance_data`) - instead, it is passed on custom app layer.
    """
    # TODO_45_75_75_65: Merge `instance_data` into `envelop_payload`:
    envelope_payload = fields.Dict(
        required = False,
    )

    @validates_schema
    def validate_known(
        self,
        input_dict: dict,
        **kwargs,
    ):
        if input_dict.get(ReservedPropName.envelope_class.name, None) == ReservedEnvelopeClass.ClassFunction.name:
            function_envelope_instance_data_desc.validate_dict(input_dict[instance_data_])

    @post_dump(pass_original = True)
    def keep_unknown_fields(
        self,
        output_dict: dict,
        orig_dict,
        **kwargs,
    ):
        """
        Dump any unknown fields in `data_envelope` as well.

        *   Because `data_envelope`-s use any (search) metadata fields, they contain arbitrary top-level keys.
        *   Because these top-level keys are arbitrary, they cannot be defined in this schema.
        *   Because they cannot be defined in the schema, they do not survive `Schema.dump`.
        This is a known issue/limitation of `marshmallow` - the `Meta.unknown` field is only used on `Schema.load`
        to allow extra keys in, but `Schema.dump` simply do not serialize them.
        This method works around it.

        See:
        https://github.com/marshmallow-code/marshmallow/issues/1545#issuecomment-947231172
        """
        for field_name in orig_dict:
            if field_name not in output_dict:
                # Do not dump `mongo_id` field because it cannot be serialized subsequently
                # (see `test_data_dump_on_server_with_non_serializable_id`):
                if field_name != mongo_id_:
                    output_dict[field_name] = orig_dict[field_name]
        return output_dict


sample_field_type_A_ = "SomeTypeA"
sample_field_type_B_ = "SomeTypeB"
sample_field_type_C_ = "SomeTypeC"

data_envelope_desc = TypeDesc(
    dict_schema = DataEnvelopeSchema(),
    ref_name = DataEnvelopeSchema.__name__,
    dict_example = {
        envelope_id_: "some_unique_id",
        instance_data_: function_envelope_instance_data_desc.dict_example,
        envelope_payload_: {},
        ReservedPropName.envelope_class.name: ReservedEnvelopeClass.ClassFunction.name,
        sample_field_type_A_: "A_value_1",
        sample_field_type_B_: "B_value_1",
        sample_field_type_C_: "C_value_1",
    },
    default_file_path = "",
)



from argrelay.misc_helper_common.ElapsedTime import ElapsedTime


def worker_main(
    call_ctx,
    client_config,
    proc_role,
    is_optimized_completion,
    w_pipe_end,
    shell_ctx,
) -> "ClientCommandAbstract":
    if client_config.use_local_requests:
        # This branch with `use_local_requests` is used only for testing
        # (to inspect internal server data via `ClientLocal` and `LocalServer`):
        from argrelay.test_infra.ClientLocal import ClientLocal
        abstract_client = ClientLocal(
            client_config,
        )
    else:
        from argrelay.relay_client.ClientRemote import ClientRemote
        abstract_client = ClientRemote(
            client_config,
            proc_role,
            w_pipe_end,
            is_optimized_completion,
        )

    ElapsedTime.measure("before_client_invocation")
    command_obj = abstract_client.make_request(call_ctx)

    ElapsedTime.measure("on_exit")
    if shell_ctx.is_debug_enabled:
        ElapsedTime.print_all()
    return command_obj



from argrelay.client_pipeline.BytesHandlerAbstract import BytesHandlerAbstract
from argrelay.relay_client.proc_spinner import child_data_chunks
from argrelay.relay_client.proc_splitter import is_child_successful, get_child_exit_code


class BytesDstReceiver:

    def __init__(
        self,
        bytes_handler: BytesHandlerAbstract,
    ):
        super().__init__(
        )
        self.bytes_handler = bytes_handler

    def receive_bytes(
        self,
    ):
        if is_child_successful():
            self.handle_bytes(b"".join(child_data_chunks))
        else:
            raise RuntimeError(f"child exit_code: {get_child_exit_code()}")

    def handle_bytes(
        self,
        given_bytes: bytes,
    ):
        self.bytes_handler.handle_bytes(given_bytes)



from argrelay.client_command_local.ClientCommandLocal import ClientCommandLocal
from argrelay.enum_desc.ProcRole import ProcRole
from argrelay.enum_desc.ServerAction import ServerAction
from argrelay.misc_helper_common.ElapsedTime import ElapsedTime
from argrelay.relay_client.ClientCommandFactoryAbstract import (
    ClientCommandFactoryAbstract,
    select_client_response_handler,
)
from argrelay.relay_server.LocalServer import LocalServer
from argrelay.runtime_data.PluginConfig import PluginConfig
from argrelay.runtime_data.ServerConfig import ServerConfig
from argrelay.schema_config_core_server.ServerConfigSchema import server_config_desc
from argrelay.schema_config_plugin.PluginConfigSchema import plugin_config_desc
from argrelay.server_spec.CallContext import CallContext


class ClientCommandFactoryLocal(ClientCommandFactoryAbstract):
    local_server: LocalServer = None
    """
    This instance of `LocalServer` allows reusing server for multiple `ClientLocal` invocation in tests.
    For example, to avoid re-loading TD_38_03_48_51 large data sets.
    """

    def __init__(self):
        self._start_local_server()

    def _start_local_server(self):
        if not ClientCommandFactoryLocal.local_server:
            self.server_config: ServerConfig = server_config_desc.obj_from_default_file()
            self.plugin_config: PluginConfig = plugin_config_desc.obj_from_default_file()
            ElapsedTime.measure("after_server_config_load")
            ClientCommandFactoryLocal.local_server = LocalServer(
                self.server_config,
                self.plugin_config,
            )
            ClientCommandFactoryLocal.local_server.start_local_server()
            ElapsedTime.measure("after_local_server_start")
        else:
            self.server_config: ServerConfig = ClientCommandFactoryLocal.local_server.server_config

    def create_command(
        self,
        call_ctx: CallContext,
    ) -> ClientCommandLocal:
        return ClientCommandLocal(
            call_ctx = call_ctx,
            client_response_handler = select_client_response_handler(
                ProcRole.SoleProcWorker,
                call_ctx.server_action,
            ),
            local_server = self.local_server,
            server_request_handler = select_local_server_request_handler(
                call_ctx.server_action,
                self.local_server,
            ),
        )


def select_local_server_request_handler(
    server_action: ServerAction,
    local_server: LocalServer,
):
    if server_action is ServerAction.ProposeArgValues:
        from argrelay.handler_request.ProposeArgValuesServerRequestHandler import (
            ProposeArgValuesServerRequestHandler,
        )
        return ProposeArgValuesServerRequestHandler(local_server)
    if server_action is ServerAction.DescribeLineArgs:
        from argrelay.handler_request.DescribeLineArgsServerRequestHandler import (
            DescribeLineArgsServerRequestHandler,
        )
        return DescribeLineArgsServerRequestHandler(local_server)
    if server_action is ServerAction.RelayLineArgs:
        from argrelay.handler_request.RelayLineArgsServerRequestHandler import (
            RelayLineArgsServerRequestHandler,
        )
        return RelayLineArgsServerRequestHandler(local_server)
    raise RuntimeError
